# encoding: utf-8

# Nagstamon - Nagios status monitor for your desktop
# Copyright (C) 2008-2026 Henri Wahl <henri@nagstamon.de> et al.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

import traceback
import urllib.request
import urllib.parse
import urllib.error
import sys
# API V2
import pprint
import json
import requests

from datetime import datetime, timedelta

from Nagstamon.objects import *
from Nagstamon.servers.Generic import GenericServer
from Nagstamon.config import conf
from Nagstamon.helpers import webbrowser_open


# This class support Centreon V2 API
# Things to do :
#  - BROWSER_URLS -> move into define_url() to be consistent

class CentreonServer(GenericServer):

    def __init__(self, **kwds):

        GenericServer.__init__(self, **kwds)

        self.TYPE = 'Centreon'

        # Centreon API uses a token
        self.token = None

        # HARD/SOFT state mapping
        self.HARD_SOFT = {'(H)': 'hard', '(S)': 'soft'}

        # Entries for monitor default actions in context menu
        # Removed Monitor, as I don’t know a way to show directly the service
        #  or host details page, so i show the main ressource page
        # self.MENU_ACTIONS = ['Monitor', 'Recheck', 'Acknowledge', 'Submit check result', 'Downtime']
        self.MENU_ACTIONS = ['Recheck', 'Acknowledge', 'Downtime']

        # URLs of the Centreon pages
        self.urls_centreon = None

        # limit number of services retrived
        self.limit_services_number = 9999

    def init_config(self):
        '''
        init_config, called at thread start
        '''
        # Version check
        result = self.fetch_url(f'{self.monitor_cgi_url}/api/latest/platform/versions', no_auth=True, giveback='raw')

        data = json.loads(result.result)
        error = result.error
        status_code = result.status_code

        # check if any error occured
        errors_occured = self.check_for_error(data, error, status_code)
        if errors_occured is not None:
            return (errors_occured)

        self.centreon_version_major = int(data["web"]["major"])
        self.centreon_version_minor = int(data["web"]["minor"])
        if conf.debug_mode is True:
            self.debug(server='[' + self.get_name() + ']', debug='Centreon version detected : ' + str(self.centreon_version_major) + '.' + str(self.centreon_version_minor))

        if self.centreon_version_major >= 21:
            # URLs for browser shortlinks/buttons on popup window
            self.BROWSER_URLS = {'monitor': '$MONITOR$/monitoring/resources',
                                 'hosts': '$MONITOR$/monitoring/resources',
                                 'services': '$MONITOR$/monitoring/resources',
                                 'history': '$MONITOR$/main.php?p=20301'}
            # RestAPI version
            if self.centreon_version_major == 21:
                self.restapi_version = "latest"
            elif self.centreon_version_major == 22:
                self.restapi_version = "v22.04"
            elif self.centreon_version_major == 23 and self.centreon_version_minor == 4:
                self.restapi_version = "v23.04"
            elif self.centreon_version_major == 23 and self.centreon_version_minor == 10:
                self.restapi_version = "v23.10"
            elif self.centreon_version_major == 24:
                self.restapi_version = "v24.04"
            else:
                self.restapi_version = "v24.04"
            if conf.debug_mode is True:
                self.debug(server='[' + self.get_name() + ']', debug='Centreon API version used : ' + self.restapi_version)

        else:
            if conf.debug_mode is True:
                self.debug(server='[' + self.get_name() + ']', debug='Unsupported Centreon version, must be >= 21')

        # Changed this because define_url was called 2 times
        if not self.tls_error and self.urls_centreon is None:
            self.define_url()

    def init_http(self):
        GenericServer.init_http(self)
        if self.session is None:
            self.session.headers.update({'Content-Type': 'application/json'})
            self.token = self.get_token().result

    def define_url(self):
        urls_centreon_api_v2 = {
            'resources': self.monitor_cgi_url + '/api/' + self.restapi_version + '/monitoring/resources',
            'login': self.monitor_cgi_url + '/api/' + self.restapi_version + '/login',
            'services': self.monitor_cgi_url + '/api/' + self.restapi_version + '/monitoring/resources',
            'hosts': self.monitor_cgi_url + '/api/' + self.restapi_version + '/monitoring/resources'
        }

        self.urls_centreon = urls_centreon_api_v2

    def open_monitor(self, host, service=''):
        # Used for self.MENU_ACTIONS = ['Monitor']
        # Autologin seems deprecated as admin must enable it globaly and use the old pages
        # Ex : http://10.66.113.52/centreon/main.php?autologin=1&useralias=admin&token=xxxxxx
        # if self.use_autologin is True:
        #     auth = '&autologin=1&useralias=' + self.username + '&token=' + self.autologin_key
        # else:
        #     auth = ''
        # webbrowser_open(self.urls_centreon['resources'] + auth )
        webbrowser_open(self.monitor_cgi_url)

    def get_token(self):
        try:
            cgi_data = {
                "security": {
                    "credentials": {
                        "login": self.username,
                        "password": self.password
                    }
                }
            }

            # Post json
            json_string = json.dumps(cgi_data)
            result = self.fetch_url(self.urls_centreon['login'], cgi_data=json_string, giveback='raw')

            data = json.loads(result.result)
            error = result.error
            status_code = result.status_code

            if conf.debug_mode:
                self.debug(server=self.get_name(),
                           debug="Fetched JSON: " + pprint.pformat(data))

            # check if any error occured
            errors_occured = self.check_for_error(data, error, status_code)
            if errors_occured is not None:
                self.debug(server=self.get_name(), debug="Authentication failed")
                return Result(result='ERROR', error="Authentication failed")

            token = data["security"]["token"]
            # ID of the user is needed by some requests
            user_id = data["contact"]["id"]

            if conf.debug_mode:
                self.debug(server='[' + self.get_name() + ']',
                           debug='API login : ' + self.username + ' / ' + self.password + ' > Token : ' + token + ' > User ID : ' + str(
                               user_id))

            self.user_id = user_id
            self.session.headers.update({'X-Auth-Token': token})
            return Result(result=token)

        except:
            traceback.print_exc(file=sys.stdout)
            result, error = self.error(sys.exc_info())
            return Result(result=result, error=error)

    def get_host(self, host):
        # https://demo.centreon.com/centreon/api/latest/monitoring/resources?page=1&limit=30&sort_by={"status_severity_code":"asc","last_status_change":"desc"}&types=["host"]&statuses=["WARNING","DOWN","CRITICAL","UNKNOWN"]
        url_hosts = self.urls_centreon['hosts'] + '?types=["host"]&search={"h.name":"' + host + '"}'

        try:
            # Get json
            result = self.fetch_url(url_hosts, giveback='raw')

            data = json.loads(result.result)
            error = result.error
            status_code = result.status_code

            # check if any error occured
            errors_occured = self.check_for_error(data, error, status_code)
            if errors_occured is not None:
                return (errors_occured)

            fqdn = str(data["result"][0]["fqdn"])

            if conf.debug_mode:
                self.debug(server='[' + self.get_name() + ']',
                           debug='Get Host FQDN or address : ' + host + " / " + fqdn)

            # Give back host or ip
            return Result(result=fqdn)

        except:
            traceback.print_exc(file=sys.stdout)
            # set checking flag back to False
            self.isChecking = False
            result, error = self.error(sys.exc_info())
            return Result(result=result, error=error)

    def get_host_and_service_id(self, host, service=''):
        if service == "":
            # Hosts only
            # https://demo.centreon.com/centreon/api/latest/monitoring/resources?page=1&limit=30&sort_by={"status_severity_code":"asc","last_status_change":"desc"}&types=["host"]&statuses=["WARNING","DOWN","CRITICAL","UNKNOWN"]
            url_hosts = self.urls_centreon['hosts'] + '?types=["host"]&search={"h.name":"' + host + '"}'

            try:
                # Get json
                result = self.fetch_url(url_hosts, giveback='raw')

                data = json.loads(result.result)
                error = result.error
                status_code = result.status_code

                # check if any error occured
                errors_occured = self.check_for_error(data, error, status_code)
                if errors_occured is not None:
                    return (errors_occured)

                host_id = data["result"][0]["id"]

                if conf.debug_mode:
                    self.debug(server='[' + self.get_name() + ']', debug='Get Host ID : ' + host + " / " + str(host_id))
                return host_id

            except:
                traceback.print_exc(file=sys.stdout)
                # set checking flag back to False
                self.isChecking = False
                result, error = self.error(sys.exc_info())
                return Result(result=result, error=error)
        else:
            # Host + Service
            if host == "Meta_Services":
                url_service = self.urls_centreon[
                                  'services'] + '?types=["metaservice"]&search={"s.name":"' + service + '"}'
            else:
                url_service = self.urls_centreon[
                                  'services'] + '?types=["service"]&search={"$and":[{"h.name":{"$eq":"' + host + '"}}, {"s.description":{"$eq":"' + service + '"}}]}'

            try:
                # Get json
                result = self.fetch_url(url_service, giveback='raw')

                data = json.loads(result.result)
                error = result.error
                status_code = result.status_code

                # check if any error occured
                errors_occured = self.check_for_error(data, error, status_code)
                if errors_occured is not None:
                    return (errors_occured)

                if host == "Meta_Services":
                    host_id = 0
                else:
                    host_id = data["result"][0]["parent"]["id"]
                service_id = data["result"][0]["id"]

                if conf.debug_mode:
                    self.debug(server='[' + self.get_name() + ']',
                               debug='Get Host / Service ID : ' + str(host_id) + " / " + str(service_id))
                return host_id, service_id

            except:
                traceback.print_exc(file=sys.stdout)
                # set checking flag back to False
                self.isChecking = False
                result, error = self.error(sys.exc_info())
                return Result(result=result, error=error)

    def get_start_end(self, host):
        # I don’t know how to get this info...
        # self.defaults_downtime_duration_hours = 2
        # self.defaults_downtime_duration_minutes = 0
        start = datetime.now()
        end = datetime.now() + timedelta(hours=2)

        return (str(start.strftime('%Y-%m-%d %H:%M')),
                str(end.strftime('%Y-%m-%d %H:%M')))

    def _get_status(self):
        '''
        Get status from Centreon Server
        '''
        # Be sure that the session is still active
        result = self.check_session()
        if result is not None:
            if result.result == 'ERROR':
                if 'urls_centreon' in result.error:
                    result.error = 'Connection error'
                return result

        # filter regexep to reduce network traffic
        # waiting to find a solution to reverse regexp
        # my first idea is
        # begin by (^(?!(.*
        # ending by )))
        # replace | by )))(^(?!(.*
        # replace )( by )|(
        self.re_service_filter = ''
        self.re_host_filter = ''
        if conf.re_service_enabled is True and conf.re_service_reverse is True:
            self.re_service_filter = '&search={"s.description":{"$rg":"' + str(conf.re_service_pattern) + '"}}'
        if conf.re_host_enabled is True and conf.re_host_reverse is True:
            self.re_host_filter = '&search={"h.name":{"$rg":"' + str(conf.re_host_pattern) + '"}}'

        # Services URL
        # https://demo.centreon.com/centreon/api/latest/monitoring/resources?page=1&limit=30&sort_by={"status_severity_code":"asc","last_status_change":"desc"}&types=["service"]&statuses=["WARNING","DOWN","CRITICAL","UNKNOWN"]
        url_services = self.urls_centreon[
                           'services'] + '?types=["metaservice","service"]&statuses=["WARNING","DOWN","CRITICAL","UNKNOWN"]' + self.re_service_filter + '&limit=' + str(
            self.limit_services_number)

        # Hosts URL
        # https://demo.centreon.com/centreon/api/latest/monitoring/resources?page=1&limit=30&sort_by={"status_severity_code":"asc","last_status_change":"desc"}&types=["host"]&statuses=["WARNING","DOWN","CRITICAL","UNKNOWN"]
        url_hosts = self.urls_centreon[
                        'hosts'] + '?types=["host"]&statuses=["WARNING","DOWN","CRITICAL","UNKNOWN"]' + self.re_host_filter + '&limit=' + str(
            self.limit_services_number)

        # Hosts
        try:
            # Get json
            result = self.fetch_url(url_hosts, giveback='raw')

            data = json.loads(result.result)
            error = result.error
            status_code = result.status_code

            # check if any error occured
            errors_occured = self.check_for_error(data, error, status_code)
            if errors_occured is not None:
                return (errors_occured)

            if data["meta"]["total"] == 0:
                self.debug(server='[' + self.get_name() + ']', debug='No host down')
            else:
                for alerts in data["result"]:
                    new_host = alerts["name"]
                    self.new_hosts[new_host] = GenericHost()
                    self.new_hosts[new_host].name = alerts["name"]
                    self.new_hosts[new_host].server = self.name
                    # API inconsistency, even by fixing exact version number, changed starting with 22.04
                    if self.centreon_version_major == 21 or (self.centreon_version_major == 22 and self.centreon_version_minor == 4):
                        self.new_hosts[new_host].criticality = alerts["severity_level"]
                    else:
                        self.new_hosts[new_host].criticality = alerts["severity"]
                    self.new_hosts[new_host].status = alerts["status"]["name"]
                    self.new_hosts[new_host].last_check = alerts["last_check"]
                    # last_state_change = datetime.strptime(alerts["last_status_change"], '%Y-%m-%dT%H:%M:%S%z').replace(tzinfo=None)
                    self.new_hosts[new_host].duration = alerts["duration"]
                    self.new_hosts[new_host].attempt = alerts["tries"]
                    self.new_hosts[new_host].status_information = alerts["information"]
                    # Change starting with 23.10
                    if (self.centreon_version_major >= 23 and self.centreon_version_minor >= 10) or self.centreon_version_major > 23:
                        self.new_hosts[new_host].passiveonly = alerts["has_passive_checks_enabled"]
                        self.new_hosts[new_host].notifications_disabled = not alerts["is_notification_enabled"]
                        self.new_hosts[new_host].acknowledged = alerts["is_acknowledged"]
                        self.new_hosts[new_host].scheduled_downtime = alerts["is_in_downtime"]
                    else:
                        self.new_hosts[new_host].passiveonly = alerts["passive_checks"]
                        self.new_hosts[new_host].notifications_disabled  = not alerts["notification_enabled"]
                        self.new_hosts[new_host].acknowledged = alerts["acknowledged"]
                        self.new_hosts[new_host].scheduled_downtime = alerts["in_downtime"]
                    # avoid crash if flapping is not configured in Centreon
                    # according to https://github.com/HenriWahl/Nagstamon/issues/866#issuecomment-1302257034
                    self.new_hosts[new_host].flapping = alerts.get("flapping", False)
                    if "(S)" in alerts["tries"]:
                        self.new_hosts[new_host].status_type = self.HARD_SOFT['(S)']
                    else:
                        self.new_hosts[new_host].status_type = self.HARD_SOFT['(H)']
                    self.debug(server='[' + self.get_name() + ']', debug='Host indexed : ' + new_host)

        except:
            traceback.print_exc(file=sys.stdout)
            # set checking flag back to False
            self.isChecking = False
            result, error = self.error(sys.exc_info())
            return Result(result=result, error=error)

        # Services
        try:
            # Get json
            result = self.fetch_url(url_services, giveback='raw')

            data = json.loads(result.result)
            error = result.error
            status_code = result.status_code

            # check if any error occured
            errors_occured = self.check_for_error(data, error, status_code)
            if errors_occured is not None:
                return (errors_occured)

            if data["meta"]["total"] == 0:
                self.debug(server='[' + self.get_name() + ']', debug='No service down')
            else:
                for alerts in data["result"]:
                    if alerts["type"] == "metaservice":
                        new_host = "Meta_Services"
                    else:
                        new_host = alerts["parent"]["name"]
                    new_service = alerts["name"]
                    # Needed if non-ok services are on a UP host
                    if not new_host in self.new_hosts:
                        self.new_hosts[new_host] = GenericHost()
                        self.new_hosts[new_host].name = new_host
                        self.new_hosts[new_host].status = 'UP'
                    self.new_hosts[new_host].services[new_service] = GenericService()
                    # Attributs à remplir
                    self.debug(server='[' + self.get_name() + ']',
                               debug='Service indexed : ' + new_host + ' / ' + new_service)

                    self.new_hosts[new_host].services[new_service].server = self.name
                    self.new_hosts[new_host].services[new_service].host = new_host
                    self.new_hosts[new_host].services[new_service].name = new_service
                    self.new_hosts[new_host].services[new_service].status = alerts["status"]["name"]
                    self.new_hosts[new_host].services[new_service].last_check = alerts["last_check"]
                    # last_state_change = datetime.strptime(alerts["last_state_change"], '%Y-%m-%dT%H:%M:%S%z').replace(tzinfo=None)
                    # self.new_hosts[new_host].services[new_service].duration = datetime.now() - last_state_change
                    self.new_hosts[new_host].services[new_service].duration = alerts["duration"]
                    self.new_hosts[new_host].services[new_service].attempt = alerts["tries"]
                    self.new_hosts[new_host].services[new_service].status_information = alerts["information"]
                    # Change starting with 23.10
                    if (self.centreon_version_major >= 23 and self.centreon_version_minor >= 10) or self.centreon_version_major > 23:
                        self.new_hosts[new_host].services[new_service].passiveonly = alerts["has_passive_checks_enabled"]
                        self.new_hosts[new_host].services[new_service].notifications_disabled = not alerts["is_notification_enabled"]
                        self.new_hosts[new_host].services[new_service].acknowledged = alerts["is_acknowledged"]
                        self.new_hosts[new_host].services[new_service].scheduled_downtime = alerts["is_in_downtime"]
                    else:
                        self.new_hosts[new_host].services[new_service].passiveonly = alerts["passive_checks"]
                        self.new_hosts[new_host].services[new_service].notifications_disabled = not alerts["notification_enabled"]
                        self.new_hosts[new_host].services[new_service].acknowledged = alerts["acknowledged"]
                        self.new_hosts[new_host].services[new_service].scheduled_downtime = alerts["in_downtime"]
                    # avoid crash if flapping is not configured in Centreon
                    # according to https://github.com/HenriWahl/Nagstamon/issues/866#issuecomment-1302257034
                    self.new_hosts[new_host].services[new_service].flapping = alerts.get("flapping", False)
                    if "(S)" in alerts["tries"]:
                        self.new_hosts[new_host].services[new_service].status_type = self.HARD_SOFT['(S)']
                    else:
                        self.new_hosts[new_host].services[new_service].status_type = self.HARD_SOFT['(H)']
                    # API inconsistency, even by fixing exact version number, changed starting with 22.04
                    if self.centreon_version_major == 21 or (self.centreon_version_major == 22 and self.centreon_version_minor == 4):
                        self.new_hosts[new_host].services[new_service].criticality = alerts["severity_level"]
                    else:
                        self.new_hosts[new_host].services[new_service].criticality = alerts["severity"]

        except:
            traceback.print_exc(file=sys.stdout)
            # set checking flag back to False
            self.isChecking = False
            result, error = self.error(sys.exc_info())
            return Result(result=result, error=error)

        # return True if all worked well
        return Result()

    def _set_acknowledge(self, host, service, author, comment, sticky, notify, persistent, all_services=None):
        try:

            acknowledgements = {
                "acknowledgement": {
                    "comment": comment,
                    "with_services": True,
                    "is_notify_contacts": notify,
                    "is_persistent_comment": persistent,
                    "is_sticky": sticky
                },
                "resources": [
                ]
            }

            # host
            if service == '':
                host_id = self.get_host_and_service_id(host)
                if self.centreon_version_major >= 24:
                    new_resource = {
                        "type": "host",
                        "id": host_id,
                        "parent": {
                            "id": host_id
                        }
                    }
                else:
                    new_resource = {
                        "type": "host",
                        "id": host_id,
                        "parent": {
                            "id": None
                        }
                    }

                acknowledgements["resources"].append(new_resource)

                # Post json
                json_string = json.dumps(acknowledgements)
                # {protocol}://{server}:{port}/centreon/api/{version}/monitoring/hosts/{host_id}/acknowledgements
                result = self.fetch_url(self.urls_centreon['resources'] + '/acknowledge', cgi_data=json_string,
                                        giveback='raw')

                error = result.error
                status_code = result.status_code

                if conf.debug_mode:
                    self.debug(server='[' + self.get_name() + ']',
                               debug="Set Ack on Host, status code : " + str(status_code))

            # Service
            if service != '' or all_services:
                if not all_services:
                    all_services = [service]

                for s in all_services:
                    host_id, service_id = self.get_host_and_service_id(host, service)

                    if host == "Meta_Services":
                        new_resource = {
                            "type": "metaservice",
                            "id": service_id,
                            "parent": {
                                "id": None
                            }
                        }

                    else:
                        new_resource = {
                            "type": "service",
                            "id": service_id,
                            "parent": {
                                "id": host_id
                            }
                        }

                    acknowledgements["resources"].append(new_resource)
                    if conf.debug_mode:
                        self.debug(server='[' + self.get_name() + ']',
                                   debug="Stack ack for Host (" + host + ") / Service (" + service + ")")

                # Post json
                json_string = json.dumps(acknowledgements)
                # {protocol}://{server}:{port}/centreon/api/{version}/monitoring/services/acknowledgements
                result = self.fetch_url(self.urls_centreon['resources'] + '/acknowledge', cgi_data=json_string,
                                        giveback='raw')

                error = result.error
                status_code = result.status_code

                if conf.debug_mode:
                    self.debug(server='[' + self.get_name() + ']',
                               debug="Set Acks, status code : " + str(status_code))

        except:
            traceback.print_exc(file=sys.stdout)
            # set checking flag back to False
            self.isChecking = False
            result, error = self.error(sys.exc_info())
            return Result(result=result, error=error)

    def _set_recheck(self, host, service):
        rechecks = {
            "resources": [
            ]
        }

        # This new parameter was added in 23.04
        if self.centreon_version_major >= 23:
            property_to_add = {
                "check": {
                    "is_forced": True
                }
            }
            rechecks.update(property_to_add)

        try:
            # Host
            if service == '':
                host_id = self.get_host_and_service_id(host)

                new_resource = {
                    "type": "host",
                    "id": host_id,
                    "parent": None
                }

                rechecks["resources"].append(new_resource)

                # Post json
                json_string = json.dumps(rechecks)
                # {protocol}://{server}:{port}/centreon/api/{version}/monitoring/resources/check
                result = self.fetch_url(self.urls_centreon['resources'] + '/check', cgi_data=json_string, giveback='raw')

                error = result.error
                status_code = result.status_code

                if conf.debug_mode:
                    self.debug(server='[' + self.get_name() + ']',
                               debug="Recheck on Host : " + host + ", status code : " + str(status_code))

            # Service
            else:
                host_id, service_id = self.get_host_and_service_id(host, service)

                if host == "Meta_Services":
                    new_resource = {
                        "type": "metaservice",
                        "id": service_id,
                        "parent": None
                    }

                else:
                    new_resource = {
                        "type": "service",
                        "id": service_id,
                        "parent": {
                            "id": host_id
                        }
                    }

                rechecks["resources"].append(new_resource)

                # Post json
                json_string = json.dumps(rechecks)
                # {protocol}://{server}:{port}/centreon/api/{version}/monitoring/resources/check
                result = self.fetch_url(self.urls_centreon['resources'] + '/check', cgi_data=json_string, giveback='raw')

                error = result.error
                status_code = result.status_code

                if conf.debug_mode:
                    self.debug(server='[' + self.get_name() + ']',
                               debug="Reckeck on Host (" + host + ") / Service (" + service + "), status code : " + str(
                                   status_code))

        except:
            traceback.print_exc(file=sys.stdout)
            # set checking flag back to False
            self.isChecking = False
            result, error = self.error(sys.exc_info())
            return Result(result=result, error=error)

    def _set_downtime(self, host, service, author, comment, fixed, start_time, end_time, hours, minutes):
        obj_start_time = datetime.strptime(start_time, '%Y-%m-%d %H:%M')
        obj_end_time = datetime.strptime(end_time, '%Y-%m-%d %H:%M')

        # Nagstamon don’t provide the TZ, we need to get it from the OS
        obj_start_time = obj_start_time.replace(tzinfo=datetime.now().astimezone().tzinfo)
        obj_end_time = obj_end_time.replace(tzinfo=datetime.now().astimezone().tzinfo)

        # duration unit is second
        duration = (hours * 3600) + (minutes * 60)

        #  API require boolean
        if fixed == 1:
            fixed = True
        else:
            fixed = False

        downtimes = {
            "downtime": {
                "comment": comment,
                "with_services": True,
                "is_fixed": fixed,
                "duration": duration,
                "start_time": obj_start_time.strftime('%Y-%m-%dT%H:%M:%S%z'),
                "end_time": obj_end_time.strftime('%Y-%m-%dT%H:%M:%S%z')
            },
            "resources": [
            ]
        }

        try:
            if service == '':
                # Host
                host_id = self.get_host_and_service_id(host)

                new_resource = {
                    "type": "host",
                    "id": host_id,
                    "parent": None
                }

                downtimes["resources"].append(new_resource)

                # Post json
                json_string = json.dumps(downtimes)
                # {protocol}://{server}:{port}/centreon/api/{version}/monitoring/resources/downtime
                result = self.fetch_url(self.urls_centreon['resources'] + '/downtime', cgi_data=json_string,
                                        giveback='raw')

                error = result.error
                status_code = result.status_code

                if conf.debug_mode:
                    self.debug(server='[' + self.get_name() + ']',
                               debug="Downtime on Host : " + host + ", status code : " + str(status_code))

            # Service
            else:
                host_id, service_id = self.get_host_and_service_id(host, service)

                if host == "Meta_Services":
                    new_resource = {
                        "type": "metaservice",
                        "id": service_id,
                        "parent": {
                            "id": None
                        }
                    }

                else:
                    new_resource = {
                        "type": "service",
                        "id": service_id,
                        "parent": {
                            "id": host_id
                        }
                    }

                downtimes["resources"].append(new_resource)

                # Post json
                json_string = json.dumps(downtimes)
                # {protocol}://{server}:{port}/centreon/api/{version}/monitoring/resources/downtime
                result = self.fetch_url(self.urls_centreon['resources'] + '/downtime', cgi_data=json_string,
                                        giveback='raw')

                error = result.error
                status_code = result.status_code

                if conf.debug_mode:
                    self.debug(server='[' + self.get_name() + ']',
                               debug="Downtime on Host (" + host + ") / Service (" + service + "), status code : " + str(
                                   status_code))


        except:
            traceback.print_exc(file=sys.stdout)
            # set checking flag back to False
            self.isChecking = False
            result, error = self.error(sys.exc_info())
            return Result(result=result, error=error)

    def check_session(self):
        if conf.debug_mode:
            self.debug(server='[' + self.get_name() + ']', debug='Checking session status')

        try:
            if conf.debug_mode:
                self.debug(server='[' + self.get_name() + ']',
                           debug='Check-session, the token expire if not been used for more than one hour. Current Token = ' + str(
                               self.token))

            cgi_data = {'limit': '1'}
            # Get en small service list, to check the status of the current token
            # This request must be done in a GET, so just encode the parameters and fetch
            result = self.fetch_url(self.urls_centreon['resources'] + '?' + urllib.parse.urlencode(cgi_data),
                                    giveback="raw")

           # If we got an 403 or 401 (and 500 for version 21.), the token expired and must be renewed
            if self.centreon_version_major == 21:
                ressources_response_list = [401, 403, 500]
            else:
                ressources_response_list = [400, 401, 403]

            if result.status_code in ressources_response_list:
                get_token_result = self.get_token()
                if get_token_result.result == 'ERROR':
                    return Result(result='ERROR', error=get_token_result.error)
                else:
                    self.token = get_token_result.result
                if conf.debug_mode:
                    self.debug(server='[' + self.get_name() + ']', debug='Check-session, session renewed')
                result = self.fetch_url(self.urls_centreon['resources'] + '?' + urllib.parse.urlencode(cgi_data),
                                        giveback="raw")
            if not 'ConnectTimeoutError' in result.error and \
                    not 'NewConnectionError' in result.error:
                data = json.loads(result.result)
                error = result.error
                status_code = result.status_code
            else:
                return Result(result='ERROR',
                              error='Connection error')

            if conf.debug_mode:
                self.debug(server=self.get_name(),
                           debug="Check-session, Fetched JSON: " + pprint.pformat(data))
                self.debug(server=self.get_name(),
                           debug="Check-session, Error : " + error + ", Status code : " + str(status_code))

        except:
            traceback.print_exc(file=sys.stdout)
            result, error = self.error(sys.exc_info())
            return Result(result=result, error=error)



# Xlib.xauth -- ~/.Xauthority access
#
#    Copyright (C) 2000 Peter Liljenberg <petli@ctrl-c.liu.se>
#    Copyright (C) 2013 LiuLang <gsushzhsosgsu@gmail.com>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,  USA

import os
import struct

from Xlib import X, error

FamilyInternet = X.FamilyInternet
FamilyDECnet = X.FamilyDECnet
FamilyChaos = X.FamilyChaos
FamilyLocal = 256

class Xauthority:
    def __init__(self, filename = None):
        if filename is None:
            filename = os.environ.get('XAUTHORITY')

        if filename is None:
            try:
                filename = os.path.join(os.environ['HOME'], '.Xauthority')
            except KeyError:
                raise error.XauthError(
                    '$HOME not set, cannot find ~/.Xauthority')

        try:
            raw = open(filename, 'rb').read()
        except OSError as err:
            raise error.XauthError('~/.Xauthority: %s' % err)

        self.entries = []

        # entry format (all shorts in big-endian)
        #   short family;
        #   short addrlen;
        #   char addr[addrlen];
        #   short numlen;
        #   char num[numlen];
        #   short namelen;
        #   char name[namelen];
        #   short datalen;
        #   char data[datalen];

        n = 0
        try:
            while n < len(raw):
                family, = struct.unpack('>H', raw[n:n+2])
                n = n + 2

                length, = struct.unpack('>H', raw[n:n+2])
                n = n + length + 2
                addr = raw[n - length : n]

                length, = struct.unpack('>H', raw[n:n+2])
                n = n + length + 2
                num = raw[n - length : n]

                length, = struct.unpack('>H', raw[n:n+2])
                n = n + length + 2
                name = raw[n - length : n]

                length, = struct.unpack('>H', raw[n:n+2])
                n = n + length + 2
                data = raw[n - length : n]

                if len(data) != length:
                    break

                self.entries.append((family, addr, num, name, data, ))
        except struct.error as e:
            print("Xlib.xauth: warning, failed to parse part of xauthority file (%s), aborting all further parsing" % filename)

        if len(self.entries) == 0:
            print("Xlib.xauth: warning, no xauthority details available")
            # raise an error?  this should get partially caught by the XNoAuthError in get_best_auth..

    def __len__(self):
        return len(self.entries)

    def __getitem__(self, i):
        return self.entries[i]

    def get_best_auth(self, family, address, dispno,
                      types = ( b"MIT-MAGIC-COOKIE-1", )):

        """Find an authentication entry matching FAMILY, ADDRESS and
        DISPNO.

        The name of the auth scheme must match one of the names in
        TYPES.  If several entries match, the first scheme in TYPES
        will be choosen.

        If an entry is found, the tuple (name, data) is returned,
        otherwise XNoAuthError is raised.
        """

        num = str(dispno).encode()
        address = address.encode()

        matches = {}

        for efam, eaddr, enum, ename, edata in self.entries:
            if efam == family and eaddr == address and num == enum:
                matches[ename] = edata

        for t in types:
            try:
                return (t, matches[t])
            except KeyError:
                pass

        raise error.XNoAuthError((family, address, dispno))



# encoding: utf-8

# Nagstamon - Nagios status monitor for your desktop
# Copyright (C) 2008-2026 Henri Wahl <henri@nagstamon.de> et al.
# Thruk additions copyright by dcec@Github
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

from collections import OrderedDict
from Nagstamon.servers.Generic import GenericServer
from Nagstamon.config import conf
import sys
import json
import datetime
import copy
import urllib.parse

from Nagstamon.helpers import human_readable_duration_from_timestamp
from Nagstamon.helpers import webbrowser_open
from Nagstamon.objects import (GenericHost, GenericService, Result)


class ThrukServer(GenericServer):
    """
        Thruk is derived from generic (Nagios) server
    """
    TYPE = 'Thruk'

    # dictionary to translate status bitmaps on webinterface into status flags
    # this are defaults from Nagios
    # "disabled.gif" is in Nagios for hosts the same as "passiveonly.gif" for services
    STATUS_MAPPING = { "ack.gif" : "acknowledged", \
                       "passiveonly.gif" : "passiveonly", \
                       "disabled.gif" : "passiveonly", \
                       "ndisabled.gif" : "notifications_disabled", \
                       "downtime.gif" : "scheduled_downtime", \
                       "flapping.gif" : "flapping"}

    # Entries for monitor default actions in context menu
    MENU_ACTIONS = ["Monitor", "Recheck", "Acknowledge", "Submit check result", "Downtime"]

    # Arguments available for submitting check results
    SUBMIT_CHECK_RESULT_ARGS = ["check_output", "performance_data"]

    # URLs for browser shortlinks/buttons on popup window
    BROWSER_URLS = { "monitor": "$MONITOR$", \
                    "hosts": "$MONITOR-CGI$/status.cgi?hostgroup=all&style=hostdetail&hoststatustypes=12&page=1&entries=all", \
                    "services": "$MONITOR-CGI$/status.cgi?dfl_s0_value_sel=5&dfl_s0_servicestatustypes=29&dfl_s0_op=%3D&style=detail&dfl_s0_type=host&dfl_s0_serviceprops=0&dfl_s0_servicestatustype=4&dfl_s0_servicestatustype=8&dfl_s0_servicestatustype=16&dfl_s0_servicestatustype=1&hidetop=&dfl_s0_hoststatustypes=15&dfl_s0_val_pre=&hidesearch=2&dfl_s0_value=all&dfl_s0_hostprops=0&nav=&page=1&entries=all", \
                    "history": "$MONITOR-CGI$/history.cgi?host=all&page=1&entries=all"}

    STATES_MAPPING = {"hosts" : {0 : "OK", 1 : "DOWN", 2 : "UNREACHABLE"}, \
                      "services" : {0 : "OK", 1 : "WARNING", 2 : "CRITICAL", 3 : "UNKNOWN"}}


    def __init__(self, **kwds):
        GenericServer.__init__(self, **kwds)


    def init_http(self):
        """
            partly not constantly working Basic Authorization requires extra Autorization headers,
            different between various server types
        """
        if self.session is None:
            GenericServer.init_http(self)

        # get cookie from login page via url retrieving as with other urls
        try:
            # login and get cookie
            if self.session is None or self.session.cookies.get('thruk_auth') is None:
                self.login()
        except:
            self.error(sys.exc_info())


    def init_config(self):
        """
            set URLs for CGI - they are static and there is no need to set them with every cycle
        """
        # create filters like described in
        # http://www.nagios-wiki.de/nagios/tips/host-_und_serviceproperties_fuer_status.cgi?s=servicestatustypes
        # Thruk allows requesting only needed information to reduce traffic
        self.cgiurl_services = self.monitor_cgi_url + "/status.cgi?host=all&servicestatustypes=28&view_mode=json&"\
                                                      "entries=all&columns=host_name,description,state,last_check,"\
                                                      "last_state_change,plugin_output,current_attempt,"\
                                                      "max_check_attempts,active_checks_enabled,is_flapping,"\
                                                      "notifications_enabled,acknowledged,state_type,"\
                                                      "scheduled_downtime_depth,host_display_name,display_name"
        # hosts (up or down or unreachable)
        self.cgiurl_hosts = self.monitor_cgi_url + "/status.cgi?hostgroup=all&style=hostdetail&"\
                                                    "dfl_s0_hoststatustypes=12&dfl_s1_hostprops=1&dfl_s2_hostprops=4&dfl_s3_hostprops=524288&&dfl_s4_hostprops=4096&dfl_s5_hostprop=16&"\
                                                    "view_mode=json&entries=all&"\
                                                    "columns=name,state,last_check,last_state_change,"\
                                                    "plugin_output,current_attempt,max_check_attempts,"\
                                                    "active_checks_enabled,notifications_enabled,is_flapping,"\
                                                    "acknowledged,scheduled_downtime_depth,state_type,host_display_name,display_name"

    def login(self):
        """
            use pure session instead of fetch_url to get Thruk session
        """
        if self.session is None:
            self.refresh_authentication = False
            GenericServer.init_http(self)

        if self.use_autologin is True:
            req = self.session.post(self.monitor_cgi_url + '/user.cgi?', data={}, headers={'X-Thruk-Auth-Key':self.autologin_key.strip()})
            if conf.debug_mode:
                self.debug(server=self.get_name(), debug='Auto Login status: ' + req.url + ' http code : ' + str(req.status_code))
            if req.status_code != 200:
                self.refresh_authentication = True
                return Result(result=None, error="Login failed")
        else:
            # set thruk test cookie to in order to directly login
            self.session.cookies.set('thruk_test', '***')
            req = self.session.post(self.monitor_cgi_url + '/login.cgi?',
                            data={'login': self.get_username(),
                                    'password': self.get_password(),
                                    'submit': 'Login'})
            if conf.debug_mode:
                self.debug(server=self.get_name(), debug='Login status: ' + req.url + ' http code : ' + str(req.status_code))
            if req.status_code != 200:
                self.refresh_authentication = True
                return Result(result=None, error="Login failed")

        if self.disabled_backends is not None:
            self.session.cookies.set('thruk_backends', '&'.join((f"{v}=2" for v in self.disabled_backends.split(','))))

    def open_monitor(self, host, service=''):
        '''
            open monitor from tablewidget context menu
        '''
        # only type is important so do not care of service '' in case of host monitor
        if service == '':
            url = self.monitor_cgi_url + '/extinfo.cgi?type=1&' + urllib.parse.urlencode( { 'host': host })
        else:
            url = self.monitor_cgi_url + '/extinfo.cgi?type=2&' + urllib.parse.urlencode( { 'host': host, 'service': self.hosts[host].services[ service ].real_name })

        if conf.debug_mode:
            self.debug(server=self.get_name(), host=host, service=service,
                       debug='Open host/service monitor web page {0}'.format(url))
        webbrowser_open(url)

    def _set_acknowledge(self, host, service, author, comment, sticky, notify, persistent, all_services=None):
        '''
            send acknowledge to monitor server - might be different on every monitor type
        '''

        url = self.monitor_cgi_url + '/cmd.cgi'

        # the following flags apply to hosts and services
        #
        # according to sf.net bug #3304098 (https://sourceforge.net/tracker/?func=detail&atid=1101370&aid=3304098&group_id=236865)
        # the send_notification-flag must not exist if it is set to 'off', otherwise
        # the Nagios core interpretes it as set, regardless its real value
        #
        # for whatever silly reason Icinga depends on the correct order of submitted form items...
        # see sf.net bug 3428844
        #
        # Thanks to Icinga ORDER OF ARGUMENTS IS IMPORTANT HERE!
        #
        cgi_data = OrderedDict()
        if service == '':
            cgi_data['cmd_typ'] = '33'
        else:
            cgi_data['cmd_typ'] = '34'
        cgi_data['cmd_mod'] = '2'
        cgi_data['host'] = host
        if service != '':
            cgi_data['service'] = self.hosts[host].services[ service ].real_name
        cgi_data['com_author'] = author
        cgi_data['com_data'] = comment
        cgi_data['btnSubmit'] = 'Commit'
        if notify is True:
            cgi_data['send_notification'] = 'on'
        if persistent is True:
            cgi_data['persistent'] = 'on'
        if sticky is True:
            cgi_data['sticky_ack'] = 'on'

        self.fetch_url(url, giveback='raw', cgi_data=cgi_data)

        # acknowledge all services on a host
        if all_services:
            for s in all_services:
                cgi_data['cmd_typ'] = '34'
                cgi_data['service'] = self.hosts[host].services[ s ].real_name
                self.fetch_url(url, giveback='raw', cgi_data=cgi_data)

    def _set_recheck(self, host, service):
        self.session.headers.update({'Content-Type': 'application/x-www-form-urlencoded; charset=UTF-8'})

        if service != '':
            if self.hosts[host].services[ service ].is_passive_only():
                # Do not check passive only checks
                return
        try:
            # get start time from Nagios as HTML to use same timezone setting like the locally installed Nagios
            result = self.fetch_url(
                self.monitor_cgi_url + '/cmd.cgi?' + urllib.parse.urlencode({'cmd_typ': '96', 'host': host}))
            self.start_time = dict(result.result.find(attrs={'name': 'start_time'}).attrs)['value']
            # decision about host or service - they have different URLs
            if service == '':
                # host
                cmd_typ = '96'
                service_name = ''
            else:
                # service @ host
                cmd_typ = '7'
                service_name = self.hosts[host].services[ service ].real_name
            # ignore empty service in case of rechecking a host
            cgi_data = urllib.parse.urlencode([('cmd_typ', cmd_typ),
                                               ('cmd_mod', '2'),
                                               ('host', host),
                                               ('service', service_name),
                                               ('start_time', self.start_time),
                                               ('force_check', 'on'),
                                               ('btnSubmit', 'Commit')])
            # execute POST request
            self.fetch_url(self.monitor_cgi_url + '/cmd.cgi', giveback='raw', cgi_data=cgi_data)
        except:
            traceback.print_exc(file=sys.stdout)

    def _set_downtime(self, host, service, author, comment, fixed, start_time, end_time, hours, minutes):
        '''
            finally send downtime command to monitor server
        '''
        url = self.monitor_cgi_url + '/cmd.cgi'

        # for some reason Icinga is very fastidiuos about the order of CGI arguments, so please
        # here we go... it took DAYS :-(
        cgi_data = OrderedDict()
        if service == '':
            cgi_data['cmd_typ'] = '55'
        else:
            cgi_data['cmd_typ'] = '56'
        cgi_data['cmd_mod'] = '2'
        cgi_data['trigger'] = '0'
        cgi_data['host'] = host
        if service != '':
            cgi_data['service'] = self.hosts[host].services[ service ].real_name
        cgi_data['com_author'] = author
        cgi_data['com_data'] = comment
        cgi_data['fixed'] = fixed
        cgi_data['start_time'] = start_time
        cgi_data['end_time'] = end_time
        cgi_data['hours'] = hours
        cgi_data['minutes'] = minutes
        cgi_data['btnSubmit'] = 'Commit'

        # running remote cgi command
        self.fetch_url(url, giveback='raw', cgi_data=cgi_data)

    def _get_status(self):
        """
            Get status from Thruk Server
        """
        # new_hosts dictionary
        self.new_hosts = dict()

        # hosts - mostly the down ones
        # unfortunately the hosts status page has a different structure so
        # hosts must be analyzed separately
        try:
            # JSON experiments
            result = self.fetch_url(self.cgiurl_hosts, giveback='raw')
            jsonraw, error, status_code = copy.deepcopy(result.result),\
                                          copy.deepcopy(result.error),\
                                          result.status_code

            # check if any error occured
            errors_occured = self.check_for_error(jsonraw, error, status_code)
            # if there are errors return them
            if errors_occured is not None:
                return errors_occured

            # in case basic auth did not work try form login cookie based login
            if jsonraw.startswith("<"):
                self.refresh_authentication = True
                return Result(result=None, error="Login failed")

            # in case JSON is not empty evaluate it
            elif not jsonraw == "[]":
                hosts = json.loads(jsonraw)

                for h in hosts:
                    if h["name"] not in self.new_hosts:
                        self.new_hosts[h["name"]] = GenericHost()
                        self.new_hosts[h["name"]].name = h["name"]
                        self.new_hosts[h["name"]].server = self.name
                        self.new_hosts[h["name"]].status = self.STATES_MAPPING["hosts"][h["state"]]
                        self.new_hosts[h["name"]].last_check = datetime.datetime.fromtimestamp(int(h["last_check"])).isoformat(" ")
                        self.new_hosts[h["name"]].duration = human_readable_duration_from_timestamp(h["last_state_change"])
                        self.new_hosts[h["name"]].attempt = "%s/%s" % (h["current_attempt"], h["max_check_attempts"])
                        self.new_hosts[h["name"]].status_information = h["plugin_output"].replace("\n", " ").strip()
                        self.new_hosts[h["name"]].passiveonly = not(bool(int(h["active_checks_enabled"])))
                        self.new_hosts[h["name"]].notifications_disabled = not(bool(int(h["notifications_enabled"])))
                        self.new_hosts[h["name"]].flapping = bool(int(h["is_flapping"]))
                        self.new_hosts[h["name"]].acknowledged = bool(int(h["acknowledged"]))
                        self.new_hosts[h["name"]].scheduled_downtime = bool(int(h["scheduled_downtime_depth"]))
                        self.new_hosts[h["name"]].status_type = {0: "soft", 1: "hard"}[h["state_type"]]
                    del h
        except:
            import traceback
            traceback.print_exc(file=sys.stdout)
            # set checking flag back to False
            self.isChecking = False
            result, error = self.error(sys.exc_info())
            return Result(result=result, error=error)

        # services
        try:
            # JSON experiments
            result = self.fetch_url(self.cgiurl_services, giveback="raw")
            jsonraw, error, status_code = copy.deepcopy(result.result),\
                                          copy.deepcopy(result.error),\
                                          result.status_code

            # check if any error occured
            errors_occured = self.check_for_error(jsonraw, error, status_code)
            # if there are errors return them
            if errors_occured is not None:
                return errors_occured

            # in case basic auth did not work try form login cookie based login
            if jsonraw.startswith("<"):
                self.refresh_authentication = True
                return Result(result=None, error="Login failed")

            # in case JSON is not empty evaluate it
            elif not jsonraw == "[]":
                services = json.loads(jsonraw)

                for s in services:
                    # host objects contain service objects
                    if s["host_name"] not in self.new_hosts:
                        self.new_hosts[s["host_name"]] = GenericHost()
                        self.new_hosts[s["host_name"]].name = s["host_name"]
                        self.new_hosts[s["host_name"]].server = self.name
                        self.new_hosts[s["host_name"]].status = "UP"

                    if self.use_display_name_service:
                        entry = s["display_name"]
                    else:
                        entry = s["description"]

                    # if a service does not exist create its object
                    if entry not in self.new_hosts[s["host_name"]].services:
                        self.new_hosts[s["host_name"]].services[ entry ] = GenericService()
                        self.new_hosts[s["host_name"]].services[ entry ].host = s["host_name"]

                        self.new_hosts[s["host_name"]].services[ entry ].name = entry
                        self.new_hosts[s["host_name"]].services[ entry ].real_name = s["description"]

                        self.new_hosts[s["host_name"]].services[ entry ].server = self.name
                        self.new_hosts[s["host_name"]].services[ entry ].status = self.STATES_MAPPING["services"][s["state"]]
                        self.new_hosts[s["host_name"]].services[ entry ].last_check = datetime.datetime.fromtimestamp(int(s["last_check"])).isoformat(" ")
                        self.new_hosts[s["host_name"]].services[ entry ].duration = human_readable_duration_from_timestamp(s["last_state_change"])
                        self.new_hosts[s["host_name"]].services[ entry ].attempt = "%s/%s" % (s["current_attempt"], s["max_check_attempts"])
                        self.new_hosts[s["host_name"]].services[ entry ].status_information = s["plugin_output"].replace("\n", " ").strip()
                        self.new_hosts[s["host_name"]].services[ entry ].passiveonly = not(bool(int(s["active_checks_enabled"])))
                        self.new_hosts[s["host_name"]].services[ entry ].notifications_disabled = not(bool(int(s["notifications_enabled"])))
                        self.new_hosts[s["host_name"]].services[ entry ].flapping = not(bool(int(s["notifications_enabled"])))
                        self.new_hosts[s["host_name"]].services[ entry ] .acknowledged = bool(int(s["acknowledged"]))
                        self.new_hosts[s["host_name"]].services[ entry ].scheduled_downtime = bool(int(s["scheduled_downtime_depth"]))
                        self.new_hosts[s["host_name"]].services[ entry ].status_type = {0: "soft", 1: "hard"}[s["state_type"]]
                        del s
        except:
            import traceback
            traceback.print_exc(file=sys.stdout)
            # set checking flag back to False
            self.isChecking = False
            result, error = self.error(sys.exc_info())
            return Result(result=result, error=error)

        # dummy return in case all is OK
        return Result()



XK_blank = 0x9df
XK_soliddiamond = 0x9e0
XK_checkerboard = 0x9e1
XK_ht = 0x9e2
XK_ff = 0x9e3
XK_cr = 0x9e4
XK_lf = 0x9e5
XK_nl = 0x9e8
XK_vt = 0x9e9
XK_lowrightcorner = 0x9ea
XK_uprightcorner = 0x9eb
XK_upleftcorner = 0x9ec
XK_lowleftcorner = 0x9ed
XK_crossinglines = 0x9ee
XK_horizlinescan1 = 0x9ef
XK_horizlinescan3 = 0x9f0
XK_horizlinescan5 = 0x9f1
XK_horizlinescan7 = 0x9f2
XK_horizlinescan9 = 0x9f3
XK_leftt = 0x9f4
XK_rightt = 0x9f5
XK_bott = 0x9f6
XK_topt = 0x9f7
XK_vertbar = 0x9f8



# Xlib.support.lock -- allocate a lock
#
#    Copyright (C) 2000 Peter Liljenberg <petli@ctrl-c.liu.se>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,  USA

class _DummyLock:
    def __init__(self):

        # This might be nerdy, but by assigning methods like this
        # instead of defining them all, we create a single bound
        # method object once instead of one each time one of the
        # methods is called.

        # This gives some speed improvements which should reduce the
        # impact of the threading infrastructure in the regular code,
        # when not using threading.

        self.acquire = self.release = self.locked = self.__noop

    def __noop(self, *args):
        return


# More optimisations: we use a single lock for all lock instances
_dummy_lock = _DummyLock()

def allocate_lock():
    return _dummy_lock



# Nagstamon - Nagios status monitor for your desktop
# Copyright (C) 2008-2026 Henri Wahl <henri@nagstamon.de> et al.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

# it is important that this import is done before importing any other qui module, because
# they may need a QApplication instance to be created
from Nagstamon.qui.widgets.app import app

from Nagstamon.qui.constants import (COLORS,
                                     COLOR_STATE_NAMES,
                                     COLOR_STATUS_LABEL,
                                     HEADERS,
                                     HEADERS_HEADERS,
                                     HEADERS_HEADERS_COLUMNS,
                                     HEADERS_HEADERS_KEYS,
                                     HEADERS_KEYS_COLUMNS,
                                     HEADERS_KEYS_HEADERS,
                                     SORT_ORDER,
                                     SORT_COLUMNS_INDEX,
                                     SPACE,
                                     WINDOW_FLAGS)
from Nagstamon.qui.globals import (dbus_connection,
                                   font,
                                   font_default,
                                   font_icons,
                                   statuswindow_properties)
from Nagstamon.qui.helpers import (check_servers,
                                   hide_macos_dock_icon)
from Nagstamon.qui.widgets.buttons import (Button,
                                           CSS_CLOSE_BUTTON,
                                           PushButtonHamburger)
from Nagstamon.qui.dialogs import dialogs
from Nagstamon.qui.dialogs.check_version import CheckVersion
from Nagstamon.qui.qt import (MediaPlayer,
                              Slot)
from Nagstamon.qui.widgets.combobox_servers import ComboBoxServers
from Nagstamon.qui.widgets.buttons import PushButtonBrowserURL
from Nagstamon.qui.widgets.draggables import (DraggableLabel,
                                              DraggableWidget)
from Nagstamon.qui.widgets.icon import QIconWithFilename
from Nagstamon.qui.widgets.labels import (ClosingLabel,
                                          LabelAllOK,
                                          ServerStatusLabel)
from Nagstamon.qui.widgets.layout import HBoxLayout
from Nagstamon.qui.widgets.mediaplayer import mediaplayer
from Nagstamon.qui.widgets.menu import (MenuAtCursor,
                                        MenuContext,
                                        MenuContextSystrayicon)
from Nagstamon.qui.widgets.model import Model
from Nagstamon.qui.widgets.nagstamon_logo import NagstamonLogo
from Nagstamon.qui.widgets.statuswindow import StatusWindow
from Nagstamon.qui.widgets.system_tray_icon import SystemTrayIcon
from Nagstamon.qui.widgets.toparea import TopArea
from Nagstamon.qui.widgets.treeview import TreeView

from Nagstamon.config import (conf,
                              OS_NON_LINUX,
                              OS,
                              OS_MACOS)

# make icon status in macOS dock accessible via NSApp, used by set_macos_dock_icon_visible()
if OS == OS_MACOS:
    from AppKit import (NSApp,
                        NSApplicationPresentationDefault,
                        NSApplicationPresentationHideDock)


# check for updates
check_version = CheckVersion()

# system tray icon
systrayicon = SystemTrayIcon()

# set to none here due to race condition
#statusw indow = None
#menu = None

# combined statusbar/status window
statuswindow = StatusWindow(dialogs=dialogs,
                            systrayicon=systrayicon)

# context menu for statuswindow etc.
menu = MenuContext(parent=statuswindow)

# necessary extra menu due to Qt5-Unity-integration
if not OS in OS_NON_LINUX:
    menu_systray = MenuContextSystrayicon(parent=statuswindow)
    menu_systray.menu_ready.connect(systrayicon.set_menu)
# menu has to be set here to solve Qt-5.10-Windows-systray-mess
# and non-existence of macOS-systray-context-menu
else:
    systrayicon.set_menu(menu)

# to be connected someday elsewhere
# server -> statuswindow remove_previous server
dialogs.server.edited_remove_previous.connect(statuswindow.remove_previous_server_vbox)
dialogs.server.create_server_vbox.connect(statuswindow.create_server_vbox)

dialogs.authentication.show_up.connect(statuswindow.hide_window)
dialogs.weblogin.show_up.connect(statuswindow.hide_window)

dialogs.settings.settings_ok.connect(statuswindow.save_position_to_conf)
# trigger the statuswindow.worker to check if debug loop is needed and if so, start it
dialogs.settings.settings_ok.connect(statuswindow.worker.debug_loop)
dialogs.settings.server_deleted.connect(statuswindow.worker.debug_loop)
dialogs.settings.changed.connect(check_servers.check)
dialogs.settings.changed.connect(statuswindow.label_all_ok.set_color)
dialogs.settings.cancelled.connect(check_servers.check)

# when there are new settings/colors recreate icons
dialogs.settings.changed.connect(systrayicon.create_icons)

# when there are new settings/colors refresh labels
dialogs.settings.changed.connect(statuswindow.statusbar.reset)

# when new settings are applied to adjust font size
dialogs.settings.changed.connect(statuswindow.statusbar.adjust_size)

# menu
dialogs.settings.changed.connect(menu.initialize)

# statuswindow
statuswindow.toparea.button_filters.clicked.connect(dialogs.settings.show_filters)
statuswindow.toparea.button_settings.clicked.connect(dialogs.settings.show)
statuswindow.toparea.action_exit.triggered.connect(statuswindow.exit)
# hide if settings dialog pops up
dialogs.settings.show_dialog.connect(statuswindow.hide_window)
# workaround for the timestamp trick to avoid flickering
dialogs.settings.show_dialog.connect(statuswindow.decrease_shown_timestamp)
# refresh all information after changed settings
dialogs.settings.changed.connect(statuswindow.refresh)
dialogs.settings.changed.connect(statuswindow.toparea.combobox_servers.fill)
# hide status window if version check finished
check_version.version_info_retrieved.connect(statuswindow.hide_window)
# start debug loop by signal
dialogs.settings.start_debug_loop.connect(statuswindow.worker.debug_loop)

# clenaup vbox after server deletion
dialogs.settings.server_deleted.connect(statuswindow.delete_server_vbox)

# reinitialize statuswindow when display mode settings were changed
dialogs.settings.changed_display_mode.connect(statuswindow.reinitialize)

# connect application exit with server missing dialog
dialogs.server_missing.window.button_exit.clicked.connect(statuswindow.exit)

# connect weblogin browser
#dialogs.weblogin.page_loaded.connect(statuswindow.refresh)

# systray connections
# show status popup when systray icon was clicked
systrayicon.show_popwin.connect(statuswindow.show_window_systrayicon)
systrayicon.hide_popwin.connect(statuswindow.hide_window)
# flashing statusicon
statuswindow.worker_notification.start_flash.connect(systrayicon.flash)
statuswindow.worker_notification.stop_flash.connect(systrayicon.reset)

# trigger showing and hiding of systray icon depending on display mode
statuswindow.systrayicon_enabled.connect(systrayicon.show)
statuswindow.systrayicon_disabled.connect(systrayicon.hide)

# retrieve systray icon position for statuswindow position calculation
statuswindow.request_systrayicon_position.connect(systrayicon.retrieve_icon_position)

# connect statuswindow to authentication dialog
statuswindow.authenticate.connect(dialogs.authentication.show_auth_dialog)

# let statuswindow show message
mediaplayer.send_message.connect(statuswindow.show_message)
# connect with statuswindow notification worker
statuswindow.worker_notification.load_sound.connect(mediaplayer.set_media)
statuswindow.worker_notification.play_sound.connect(mediaplayer.play)

# necessary extra menu due to Qt5-Unity-integration
if not OS in OS_NON_LINUX:
    # change menu if there are changes in settings/servers
    dialogs.settings.changed.connect(menu_systray.initialize)
    menu_systray.action_settings.triggered.connect(statuswindow.hide_window)
    menu_systray.action_settings.triggered.connect(dialogs.settings.show)
    menu_systray.action_save_position.triggered.connect(statuswindow.save_position_to_conf)
    menu_systray.action_about.triggered.connect(statuswindow.hide_window)
    menu_systray.action_about.triggered.connect(dialogs.about.show)
    menu_systray.menu_ready.connect(systrayicon.set_menu)
    menu_systray.menu_ready.emit(menu_systray)

# needs to be emitted adter signal/slots are connected,
# might be necessary for others too
if conf.icon_in_systray:
    statuswindow.systrayicon_enabled.emit()
else:
    statuswindow.systrayicon_disabled.emit()
# tell the widgets that the menu is ready
menu.menu_ready.emit(menu)



# Xlib.protocol.__init__ -- glue for Xlib.protocol package
#
#    Copyright (C) 2000 Peter Liljenberg <petli@ctrl-c.liu.se>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,  USA

__all__ = [
    'display',
    'event',
    'request',
    'rq',
    'structs',
    ]



# Nagstamon - Nagios status monitor for your desktop
# Copyright (C) 2008-2026 Henri Wahl <henri@nagstamon.de> et al.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

from platform import release
from os import sep
from sys import argv

from Nagstamon.config import (OS,
                              OS_WINDOWS,
                              RESOURCES)
from Nagstamon.qui.qt import (Qt,
                              QApplication,
                              QFontDatabase,
                              QT_VERSION_MAJOR)
from Nagstamon.qui.widgets.buttons import StandardDialogIconsProxyStyle

# since Qt6 HighDPI-awareness is default behaviour
if QT_VERSION_MAJOR < 6:
    # enable HighDPI-awareness to avoid https://github.com/HenriWahl/Nagstamon/issues/618
    try:
        QApplication.setAttribute(Qt.AA_EnableHighDpiScaling, True)
    except AttributeError:
        pass
    QApplication.setAttribute(Qt.AA_UseHighDpiPixmaps, True)


# global application instance
app = QApplication(argv)

# set style for tooltips globally - to sad not all properties can be set here
app.setStyleSheet('''QToolTip { margin: 3px;
                                }''')

# proxy style helps to remove ugly icons in Linux Fusion style
app.setStyle(StandardDialogIconsProxyStyle())

# as long as Windows 11 + Qt6 looks that ugly it's better to choose another app style
# might be mitigated with sometimes, so commented out now
if OS == OS_WINDOWS and release() >= '11':
    app.setStyle('fusion')

# add nagstamon.ttf with icons to fonts
QFontDatabase.addApplicationFont(f'{RESOURCES}{sep}nagstamon.ttf')


# Xlib.xobject.drawable -- drawable objects (window and pixmap)
#
#    Copyright (C) 2000 Peter Liljenberg <petli@ctrl-c.liu.se>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,  USA


from Xlib import X, Xatom, Xutil
from Xlib.protocol import request, rq

# Other X resource objects
from Xlib.xobject import resource, colormap, cursor, fontable, icccm

class Drawable(resource.Resource):
    __drawable__ = resource.Resource.__resource__

    def get_geometry(self):
        return request.GetGeometry(display = self.display,
                                   drawable = self)

    def create_pixmap(self, width, height, depth):
        pid = self.display.allocate_resource_id()
        request.CreatePixmap(display = self.display,
                             depth = depth,
                             pid = pid,
                             drawable = self.id,
                             width = width,
                             height = height)

        cls = self.display.get_resource_class('pixmap', Pixmap)
        return cls(self.display, pid, owner = 1)

    def create_gc(self, **keys):
        cid = self.display.allocate_resource_id()
        request.CreateGC(display = self.display,
                         cid = cid,
                         drawable = self.id,
                         attrs = keys)

        cls = self.display.get_resource_class('gc', fontable.GC)
        return cls(self.display, cid, owner = 1)

    def copy_area(self, gc, src_drawable, src_x, src_y, width, height, dst_x, dst_y, onerror = None):
        request.CopyArea(display = self.display,
                         onerror = onerror,
                         src_drawable = src_drawable,
                         dst_drawable = self.id,
                         gc = gc,
                         src_x = src_x,
                         src_y = src_y,
                         dst_x = dst_x,
                         dst_y = dst_y,
                         width = width,
                         height = height)

    def copy_plane(self, gc, src_drawable, src_x, src_y, width, height,
                   dst_x, dst_y, bit_plane, onerror = None):
        request.CopyPlane(display = self.display,
                          onerror = onerror,
                          src_drawable = src_drawable,
                          dst_drawable = self.id,
                          gc = gc,
                          src_x = src_x,
                          src_y = src_y,
                          dst_x = dst_x,
                          dst_y = dst_y,
                          width = width,
                          height = height,
                          bit_plane = bit_plane)

    def poly_point(self, gc, coord_mode, points, onerror = None):
        request.PolyPoint(display = self.display,
                          onerror = onerror,
                          coord_mode = coord_mode,
                          drawable = self.id,
                          gc = gc,
                          points = points)

    def point(self, gc, x, y, onerror = None):
        request.PolyPoint(display = self.display,
                          onerror = onerror,
                          coord_mode = X.CoordModeOrigin,
                          drawable = self.id,
                          gc = gc,
                          points = [(x, y)])

    def poly_line(self, gc, coord_mode, points, onerror = None):
        request.PolyLine(display = self.display,
                         onerror = onerror,
                         coord_mode = coord_mode,
                         drawable = self.id,
                         gc = gc,
                         points = points)

    def line(self, gc, x1, y1, x2, y2, onerror = None):
        request.PolySegment(display = self.display,
                            onerror = onerror,
                            drawable = self.id,
                            gc = gc,
                            segments = [(x1, y1, x2, y2)])

    def poly_segment(self, gc, segments, onerror = None):
        request.PolySegment(display = self.display,
                            onerror = onerror,
                            drawable = self.id,
                            gc = gc,
                            segments = segments)

    def poly_rectangle(self, gc, rectangles, onerror = None):
        request.PolyRectangle(display = self.display,
                              onerror = onerror,
                              drawable = self.id,
                              gc = gc,
                              rectangles = rectangles)

    def rectangle(self, gc, x, y, width, height, onerror = None):
        request.PolyRectangle(display = self.display,
                              onerror = onerror,
                              drawable = self.id,
                              gc = gc,
                              rectangles = [(x, y, width, height)])


    def poly_arc(self, gc, arcs, onerror = None):
        request.PolyArc(display = self.display,
                        onerror = onerror,
                        drawable = self.id,
                        gc = gc,
                        arcs = arcs)

    def arc(self, gc,  x, y, width, height, angle1, angle2, onerror = None):
        request.PolyArc(display = self.display,
                        onerror = onerror,
                        drawable = self.id,
                        gc = gc,
                        arcs = [(x, y, width, height, angle1, angle2)])

    def fill_poly(self, gc, shape, coord_mode, points, onerror = None):
        request.FillPoly(display = self.display,
                         onerror = onerror,
                         shape = shape,
                         coord_mode = coord_mode,
                         drawable = self.id,
                         gc = gc,
                         points = points)

    def poly_fill_rectangle(self, gc, rectangles, onerror = None):
        request.PolyFillRectangle(display = self.display,
                                  onerror = onerror,
                                  drawable = self.id,
                                  gc = gc,
                                  rectangles = rectangles)

    def fill_rectangle(self, gc, x, y, width, height, onerror = None):
        request.PolyFillRectangle(display = self.display,
                                  onerror = onerror,
                                  drawable = self.id,
                                  gc = gc,
                                  rectangles = [(x, y, width, height)])

    def poly_fill_arc(self, gc, arcs, onerror = None):
        request.PolyFillArc(display = self.display,
                            onerror = onerror,
                            drawable = self.id,
                            gc = gc,
                            arcs = arcs)

    def fill_arc(self, gc,  x, y, width, height, angle1, angle2, onerror = None):
        request.PolyFillArc(display = self.display,
                            onerror = onerror,
                            drawable = self.id,
                            gc = gc,
                            arcs = [(x, y, width, height, angle1, angle2)])


    def put_image(self, gc, x, y, width, height, format,
                  depth, left_pad, data, onerror = None):
        request.PutImage(display = self.display,
                         onerror = onerror,
                         format = format,
                         drawable = self.id,
                         gc = gc,
                         width = width,
                         height = height,
                         dst_x = x,
                         dst_y = y,
                         left_pad = left_pad,
                         depth = depth,
                         data = data)

    # Trivial little method for putting PIL images.  Will break on anything
    # but depth 1 or 24...
    def put_pil_image(self, gc, x, y, image, onerror = None):
        width, height = image.size
        if image.mode == '1':
            format = X.XYBitmap
            depth = 1
            if self.display.info.bitmap_format_bit_order == 0:
                rawmode = '1;R'
            else:
                rawmode = '1'
            pad = self.display.info.bitmap_format_scanline_pad
            stride = roundup(width, pad) >> 3
        elif image.mode == 'RGB':
            format = X.ZPixmap
            depth = 24
            if self.display.info.image_byte_order == 0:
                rawmode = 'BGRX'
            else:
                rawmode = 'RGBX'
            pad = self.display.info.bitmap_format_scanline_pad
            unit = self.display.info.bitmap_format_scanline_unit
            stride = roundup(width * unit, pad) >> 3
        else:
            raise ValueError('Unknown data format')

        maxlen = (self.display.info.max_request_length << 2) \
                 - request.PutImage._request.static_size
        split = maxlen // stride

        x1 = 0
        x2 = width
        y1 = 0

        while y1 < height:
            h = min(height, split)
            if h < height:
                subimage = image.crop((x1, y1, x2, y1 + h))
            else:
                subimage = image
            w, h = subimage.size
            data = subimage.tostring("raw", rawmode, stride, 0)
            self.put_image(gc, x, y, w, h, format, depth, 0, data)
            y1 = y1 + h
            y = y + h


    def get_image(self, x, y, width, height, format, plane_mask):
        return request.GetImage(display = self.display,
                                format = format,
                                drawable = self.id,
                                x = x,
                                y = y,
                                width = width,
                                height = height,
                                plane_mask = plane_mask)

    def draw_text(self, gc, x, y, text, onerror = None):
        request.PolyText8(display = self.display,
                          onerror = onerror,
                          drawable = self.id,
                          gc = gc,
                          x = x,
                          y = y,
                          items = [text])

    def poly_text(self, gc, x, y, items, onerror = None):
        request.PolyText8(display = self.display,
                          onerror = onerror,
                          drawable = self.id,
                          gc = gc,
                          x = x,
                          y = y,
                          items = items)

    def poly_text_16(self, gc, x, y, items, onerror = None):
        request.PolyText16(display = self.display,
                           onerror = onerror,
                           drawable = self.id,
                           gc = gc,
                           x = x,
                           y = y,
                           items = items)

    def image_text(self, gc, x, y, string, onerror = None):
        request.ImageText8(display = self.display,
                           onerror = onerror,
                           drawable = self.id,
                           gc = gc,
                           x = x,
                           y = y,
                           string = string)

    def image_text_16(self, gc, x, y, string, onerror = None):
        request.ImageText16(display = self.display,
                            onerror = onerror,
                            drawable = self.id,
                            gc = gc,
                            x = x,
                            y = y,
                            string = string)

    def query_best_size(self, item_class, width, height):
        return request.QueryBestSize(display = self.display,
                                     item_class = item_class,
                                     drawable = self.id,
                                     width = width,
                                     height = height)

class Window(Drawable):
    __window__ = resource.Resource.__resource__

    def create_window(self, x, y, width, height, border_width, depth,
                      window_class =  X.CopyFromParent,
                      visual = X.CopyFromParent,
                      onerror = None,
                      **keys):

        wid = self.display.allocate_resource_id()
        request.CreateWindow(display = self.display,
                             onerror = onerror,
                             depth = depth,
                             wid = wid,
                             parent = self.id,
                             x = x,
                             y = y,
                             width = width,
                             height = height,
                             border_width = border_width,
                             window_class = window_class,
                             visual = visual,
                             attrs = keys)

        cls = self.display.get_resource_class('window', Window)
        return cls(self.display, wid, owner = 1)

    def change_attributes(self, onerror = None, **keys):
        request.ChangeWindowAttributes(display = self.display,
                                       onerror = onerror,
                                       window = self.id,
                                       attrs = keys)

    def get_attributes(self):
        return request.GetWindowAttributes(display = self.display,
                                           window = self.id)

    def destroy(self, onerror = None):
        request.DestroyWindow(display = self.display,
                              onerror = onerror,
                              window = self.id)

        self.display.free_resource_id(self.id)

    def destroy_sub_windows(self, onerror = None):
        request.DestroySubWindows(display = self.display,
                                  onerror = onerror,
                                  window = self.id)


    def change_save_set(self, mode, onerror = None):
        request.ChangeSaveSet(display = self.display,
                              onerror = onerror,
                              mode = mode,
                              window = self.id)

    def reparent(self, parent, x, y, onerror = None):
        request.ReparentWindow(display = self.display,
                               onerror = onerror,
                               window = self.id,
                               parent = parent,
                               x = x,
                               y = y)

    def map(self, onerror = None):
        request.MapWindow(display = self.display,
                          onerror = onerror,
                          window = self.id)

    def map_sub_windows(self, onerror = None):
        request.MapSubwindows(display = self.display,
                              onerror = onerror,
                              window = self.id)

    def unmap(self, onerror = None):
        request.UnmapWindow(display = self.display,
                            onerror = onerror,
                            window = self.id)

    def unmap_sub_windows(self, onerror = None):
        request.UnmapSubwindows(display = self.display,
                                onerror = onerror,
                                window = self.id)

    def configure(self, onerror = None, **keys):
        request.ConfigureWindow(display = self.display,
                                onerror = onerror,
                                window = self.id,
                                attrs = keys)

    def circulate(self, direction, onerror = None):
        request.CirculateWindow(display = self.display,
                                onerror = onerror,
                                direction = direction,
                                window = self.id)

    def raise_window(self, onerror = None):
        """alias for raising the window to the top - as in XRaiseWindow"""
        self.configure(onerror, stack_mode = X.Above)

    def query_tree(self):
        return request.QueryTree(display = self.display,
                                 window = self.id)


    def change_property(self, property, type, format, data,
                        mode = X.PropModeReplace, onerror = None):

        request.ChangeProperty(display = self.display,
                               onerror = onerror,
                               mode = mode,
                               window = self.id,
                               property = property,
                               type = type,
                               data = (format, data))

    def delete_property(self, property, onerror = None):
        request.DeleteProperty(display = self.display,
                               onerror = onerror,
                               window = self.id,
                               property = property)

    def get_property(self, property, type, offset, length, delete = 0):
        r = request.GetProperty(display = self.display,
                                delete = delete,
                                window = self.id,
                                property = property,
                                type = type,
                                long_offset = offset,
                                long_length = length)

        if r.property_type:
            fmt, value = r.value
            r.format = fmt
            r.value = value
            return r
        else:
            return None

    def get_full_property(self, property, type, sizehint = 10):
        prop = self.get_property(property, type, 0, sizehint)
        if prop:
            val = prop.value
            if prop.bytes_after:
                prop = self.get_property(property, type, sizehint,
                                         prop.bytes_after // 4 + 1)
                val = val + prop.value

            prop.value = val
            return prop
        else:
            return None

    def list_properties(self):
        r = request.ListProperties(display = self.display,
                                   window = self.id)
        return r.atoms

    def set_selection_owner(self, selection, time, onerror = None):
        request.SetSelectionOwner(display = self.display,
                                  onerror = onerror,
                                  window = self.id,
                                  selection = selection,
                                  time = time)

    def convert_selection(self, selection, target, property, time, onerror = None):
        request.ConvertSelection(display = self.display,
                                 onerror = onerror,
                                 requestor = self.id,
                                 selection = selection,
                                 target = target,
                                 property = property,
                                 time = time)

    def send_event(self, event, event_mask = 0, propagate = 0, onerror = None):
        request.SendEvent(display = self.display,
                          onerror = onerror,
                          propagate = propagate,
                          destination = self.id,
                          event_mask = event_mask,
                          event = event)

    def grab_pointer(self, owner_events, event_mask,
                     pointer_mode, keyboard_mode,
                     confine_to, cursor, time):

        r = request.GrabPointer(display = self.display,
                                owner_events = owner_events,
                                grab_window = self.id,
                                event_mask = event_mask,
                                pointer_mode = pointer_mode,
                                keyboard_mode = keyboard_mode,
                                confine_to = confine_to,
                                cursor = cursor,
                                time = time)
        return r.status

    def grab_button(self, button, modifiers, owner_events, event_mask,
                    pointer_mode, keyboard_mode,
                    confine_to, cursor, onerror = None):

        request.GrabButton(display = self.display,
                           onerror = onerror,
                           owner_events = owner_events,
                           grab_window = self.id,
                           event_mask = event_mask,
                           pointer_mode = pointer_mode,
                           keyboard_mode = keyboard_mode,
                           confine_to = confine_to,
                           cursor = cursor,
                           button = button,
                           modifiers = modifiers)

    def ungrab_button(self, button, modifiers, onerror = None):
        request.UngrabButton(display = self.display,
                             onerror = onerror,
                             button = button,
                             grab_window = self.id,
                             modifiers = modifiers)


    def grab_keyboard(self, owner_events, pointer_mode, keyboard_mode, time):
        r = request.GrabKeyboard(display = self.display,
                                 owner_events = owner_events,
                                 grab_window = self.id,
                                 time = time,
                                 pointer_mode = pointer_mode,
                                 keyboard_mode = keyboard_mode)

        return r.status

    def grab_key(self, key, modifiers, owner_events, pointer_mode, keyboard_mode, onerror = None):
        request.GrabKey(display = self.display,
                        onerror = onerror,
                        owner_events = owner_events,
                        grab_window = self.id,
                        modifiers = modifiers,
                        key = key,
                        pointer_mode = pointer_mode,
                        keyboard_mode = keyboard_mode)

    def ungrab_key(self, key, modifiers, onerror = None):
        request.UngrabKey(display = self.display,
                          onerror = onerror,
                          key = key,
                          grab_window = self.id,
                          modifiers = modifiers)

    def query_pointer(self):
        return request.QueryPointer(display = self.display,
                                    window = self.id)

    def get_motion_events(self, start, stop):
        r = request.GetMotionEvents(display = self.display,
                                    window = self.id,
                                    start = start,
                                    stop = stop)
        return r.events

    def translate_coords(self, src_window, src_x, src_y):
        return request.TranslateCoords(display = self.display,
                                       src_wid = src_window,
                                       dst_wid = self.id,
                                       src_x = src_x,
                                       src_y = src_y)

    def warp_pointer(self, x, y, src_window = 0, src_x = 0, src_y = 0,
                     src_width = 0, src_height = 0, onerror = None):

        request.WarpPointer(display = self.display,
                            onerror = onerror,
                            src_window = src_window,
                            dst_window = self.id,
                            src_x = src_x,
                            src_y = src_y,
                            src_width = src_width,
                            src_height = src_height,
                            dst_x = x,
                            dst_y = y)

    def set_input_focus(self, revert_to, time, onerror = None):
        request.SetInputFocus(display = self.display,
                              onerror = onerror,
                              revert_to = revert_to,
                              focus = self.id,
                              time = time)

    def clear_area(self, x = 0, y = 0, width = 0, height = 0, exposures = 0, onerror = None):
        request.ClearArea(display = self.display,
                          onerror = onerror,
                          exposures = exposures,
                          window = self.id,
                          x = x,
                          y = y,
                          width = width,
                          height = height)

    def create_colormap(self, visual, alloc):
        mid = self.display.allocate_resource_id()
        request.CreateColormap(display = self.display,
                               alloc = alloc,
                               mid = mid,
                               window = self.id,
                               visual = visual)
        cls = self.display.get_resource_class('colormap', colormap.Colormap)
        return cls(self.display, mid, owner = 1)

    def list_installed_colormaps(self):
        r = request.ListInstalledColormaps(display = self.display,
                                           window = self.id)
        return r.cmaps

    def rotate_properties(self, properties, delta, onerror = None):
        request.RotateProperties(display = self.display,
                                 onerror = onerror,
                                 window = self.id,
                                 delta = delta,
                                 properties = properties)

    def set_wm_name(self, name, onerror = None):
        self.change_property(Xatom.WM_NAME, Xatom.STRING, 8, name,
                             onerror = onerror)

    def get_wm_name(self):
        d = self.get_full_property(Xatom.WM_NAME, Xatom.STRING)
        if d is None or d.format != 8:
            return None
        else:
            return d.value

    def set_wm_icon_name(self, name, onerror = None):
        self.change_property(Xatom.WM_ICON_NAME, Xatom.STRING, 8, name,
                             onerror = onerror)

    def get_wm_icon_name(self):
        d = self.get_full_property(Xatom.WM_ICON_NAME, Xatom.STRING)
        if d is None or d.format != 8:
            return None
        else:
            return d.value


    def set_wm_class(self, inst, cls, onerror = None):
        self.change_property(Xatom.WM_CLASS, Xatom.STRING, 8,
                             '%s\0%s\0' % (inst, cls),
                             onerror = onerror)

    def get_wm_class(self):
        d = self.get_full_property(Xatom.WM_CLASS, Xatom.STRING)
        if d is None or d.format != 8:
            return None
        else:
            parts = d.value.split('\0')
            if len(parts) < 2:
                return None
            else:
                return parts[0], parts[1]

    def set_wm_transient_for(self, window, onerror = None):
        self.change_property(Xatom.WM_TRANSIENT_FOR, Xatom.WINDOW,
                             32, window.id,
                             onerror = onerror)

    def get_wm_transient_for(self):
        d = self.get_property(Xatom.WM_TRANSIENT_FOR, Xatom.WINDOW, 0, 1)
        if d is None or d.format != 32 or len(d.value) < 1:
            return None
        else:
            cls = self.display.get_resource_class('window', Window)
            return cls(self.display, d.value[0])


    def set_wm_protocols(self, protocols, onerror = None):
        self.change_property(self.display.get_atom('WM_PROTOCOLS'),
                             Xatom.ATOM, 32, protocols,
                             onerror = onerror)

    def get_wm_protocols(self):
        d = self.get_full_property(self.display.get_atom('WM_PROTOCOLS'), Xatom.ATOM)
        if d is None or d.format != 32:
            return []
        else:
            return d.value

    def set_wm_colormap_windows(self, windows, onerror = None):
        self.change_property(self.display.get_atom('WM_COLORMAP_WINDOWS'),
                             Xatom.WINDOW, 32,
                             [w.id for w in windows],
                             onerror = onerror)

    def get_wm_colormap_windows(self):
        d = self.get_full_property(self.display.get_atom('WM_COLORMAP_WINDOWS'),
                                   Xatom.WINDOW)
        if d is None or d.format != 32:
            return []
        else:
            cls = self.display.get_resource_class('window', Window)
            return list(map(lambda i, d = self.display, c = cls: c(d, i),
                       d.value))


    def set_wm_client_machine(self, name, onerror = None):
        self.change_property(Xatom.WM_CLIENT_MACHINE, Xatom.STRING, 8, name,
                             onerror = onerror)

    def get_wm_client_machine(self):
        d = self.get_full_property(Xatom.WM_CLIENT_MACHINE, Xatom.STRING)
        if d is None or d.format != 8:
            return None
        else:
            return d.value

    def set_wm_normal_hints(self, hints = {}, onerror = None, **keys):
        self._set_struct_prop(Xatom.WM_NORMAL_HINTS, Xatom.WM_SIZE_HINTS,
                              icccm.WMNormalHints, hints, keys, onerror)

    def get_wm_normal_hints(self):
        return self._get_struct_prop(Xatom.WM_NORMAL_HINTS, Xatom.WM_SIZE_HINTS,
                                     icccm.WMNormalHints)

    def set_wm_hints(self, hints = {}, onerror = None, **keys):
        self._set_struct_prop(Xatom.WM_HINTS, Xatom.WM_HINTS,
                              icccm.WMHints, hints, keys, onerror)

    def get_wm_hints(self):
        return self._get_struct_prop(Xatom.WM_HINTS, Xatom.WM_HINTS,
                                     icccm.WMHints)

    def set_wm_state(self, hints = {}, onerror = None, **keys):
        atom = self.display.get_atom('WM_STATE')
        self._set_struct_prop(atom, atom, icccm.WMState, hints, keys, onerror)

    def get_wm_state(self):
        atom = self.display.get_atom('WM_STATE')
        return self._get_struct_prop(atom, atom, icccm.WMState)

    def set_wm_icon_size(self, hints = {}, onerror = None, **keys):
        self._set_struct_prop(Xatom.WM_ICON_SIZE, Xatom.WM_ICON_SIZE,
                              icccm.WMIconSize, hints, keys, onerror)

    def get_wm_icon_size(self):
        return self._get_struct_prop(Xatom.WM_ICON_SIZE, Xatom.WM_ICON_SIZE,
                                     icccm.WMIconSize)

    # Helper function for getting structured properties.
    # pname and ptype are atoms, and pstruct is a Struct object.
    # Returns a DictWrapper, or None

    def _get_struct_prop(self, pname, ptype, pstruct):
        r = self.get_property(pname, ptype, 0, pstruct.static_size // 4)
        if r and r.format == 32:
            value = r.value.tostring()
            if len(value) == pstruct.static_size:
                return pstruct.parse_binary(value, self.display)[0]

        return None

    # Helper function for setting structured properties.
    # pname and ptype are atoms, and pstruct is a Struct object.
    # hints is a mapping or a DictWrapper, keys is a mapping.  keys
    # will be modified.  onerror is the error handler.

    def _set_struct_prop(self, pname, ptype, pstruct, hints, keys, onerror):
        if isinstance(hints, rq.DictWrapper):
            keys.update(hints._data)
        else:
            keys.update(hints)

        value = pstruct.to_binary(*(), **keys)

        self.change_property(pname, ptype, 32, value, onerror = onerror)


class Pixmap(Drawable):
    __pixmap__ = resource.Resource.__resource__

    def free(self, onerror = None):
        request.FreePixmap(display = self.display,
                           onerror = onerror,
                           pixmap = self.id)

        self.display.free_resource_id(self.id)

    def create_cursor(self, mask,
                      f_rgb, b_rgb,
                      x, y):
        fore_red, fore_green, fore_blue = f_rgb
        back_red, back_green, back_blue = b_rgb
        cid = self.display.allocate_resource_id()
        request.CreateCursor(display = self.display,
                             cid = cid,
                             source = self.id,
                             mask = mask,
                             fore_red = fore_red,
                             fore_green = fore_green,
                             fore_blue = fore_blue,
                             back_red = back_red,
                             back_green = back_green,
                             back_blue = back_blue,
                             x = x,
                             y = y)
        cls = self.display.get_resource_class('cursor', cursor.Cursor)
        return cls(self.display, cid, owner = 1)


def roundup(value, unit):
    return (value + (unit - 1)) & ~(unit - 1)



# -*- encoding: utf-8; py-indent-offset: 4 -*-
#
# Zabbix.py based on Checkmk Multisite.py
import base64
import json
import sys
import time
import datetime
import socket
from packaging import version

from Nagstamon.helpers import (human_readable_duration_from_timestamp,
                               webbrowser_open)
from Nagstamon.config import conf
from Nagstamon.objects import (GenericHost,
                               GenericService,
                               Result)
from Nagstamon.servers.Generic import GenericServer, BearerAuth


class ZabbixError(Exception):

    def __init__(self, terminate, result):
        self.terminate = terminate
        self.result = result


class ZabbixServer(GenericServer):
    """
       special treatment for Zabbix, taken from Check_MK Multisite JSON API
    """
    TYPE = 'Zabbix'

    def __init__(self, **kwds):
        GenericServer.__init__(self, **kwds)

        # Prepare all urls needed by nagstamon -
        self.authentication = conf.servers[self.get_name()].authentication
        self.urls = {}
        # self.statemap = {}
        self.statemap = {
            'UNREACH': 'UNREACHABLE',
            'CRIT': 'CRITICAL',
            'WARN': 'WARNING',
            'UNKN': 'UNKNOWN',
            'PEND': 'PENDING',
            '0': 'OK',
            '1': 'INFORMATION',
            '2': 'WARNING',
            '3': 'AVERAGE',
            '4': 'HIGH',
            '5': 'DISASTER'}

        # Entries for monitor default actions in context menu
        self.MENU_ACTIONS = ['Monitor', 'Acknowledge', 'Downtime']
        # URLs for browser shortlinks/buttons on popup window
        self.BROWSER_URLS = {'monitor': '$MONITOR$',
                             'hosts': '$MONITOR$/zabbix.php?action=host.view',
                             'services': '$MONITOR$/zabbix.php?action=problem.view&fullscreen=0&page=1&filter_show=3&filter_set=1',
                             'history': '$MONITOR$/zabbix.php?action=problem.view&fullscreen=0&page=1&filter_show=2&filter_set=1'}

        self.username = conf.servers[self.get_name()].username
        self.password = conf.servers[self.get_name()].password
        self.timeout = conf.servers[self.get_name()].timeout
        self.ignore_cert = conf.servers[self.get_name()].ignore_cert
        self.use_description_name_service = conf.servers[self.get_name()].use_description_name_service
        self.api_version = ''
        self.auth_token = ''
        # Force authentication refresh by default until verified
        self.refresh_authentication = True
        self.monitor_path = '/api_jsonrpc.php'

    def init_config(self):
        super().init_config()
        self.monitor_url = self.monitor_url.split(self.monitor_path)[0]
        self.monitor_cgi_url = f"{self.monitor_url}{self.monitor_path}"

    def init_http(self):
        """
        things to do if HTTP is not initialized
        Ensure a valid HTTP session exists before using it and handle re-auth flows.
        """
        # Let GenericServer manage refresh/session lifecycle
        super().init_http()

        # If refresh was requested, GenericServer cleared the session and returned False.
        # Create a fresh session explicitly so we can proceed with headers, version and auth checks.
        if self.session is None:
            self.session = self.create_session()

        # prepare for JSON
        if not hasattr(self.session, 'headers') or self.session.headers is None:
            # extremely defensive, but keeps AttributeError away
            self.session.headers = {}
        self.session.headers.update({'Accept': 'application/json',
                                     'Content-Type': 'application/json-rpc'})
        try:
            self.set_zabbix_version()
            if not self.api_version:
                raise RuntimeError('Failed to determine Zabbix API version. Please verify the server URL and API endpoint are correct. Also check your network connectivity and authentication credentials.')
            self.check_authentication()
            if self.refresh_authentication:
                self.login()
        except Exception:
            self.error(sys.exc_info())
            return

    def api_request(self, cgi_data, no_auth=False):
        """
            Make a request to the Zabbix API
            Returns the response as a dictionary
        """
        url = self.monitor_cgi_url
        result = self.fetch_url(url,
                                  headers=self.session.headers,
                                  cgi_data=cgi_data,
                                  giveback='json',
                                  no_auth=no_auth)
        # Check if the result is a valid JSON response
        data = result.result
        error = result.error
        status_code = result.status_code
        if error:
            raise ZabbixError(terminate=True, result=Result(result=False, error=error, status_code=status_code))
        return data

    def login(self):
        if conf.servers[self.get_name()].authentication == 'bearer':
            # In bearer mode the GUI stores the API token in the password field
            token = self.password
            if token:
                self.session.auth = BearerAuth(token)
                self.refresh_authentication = False
                return

        # check version to use the correct keyword for username which changed since 6.4
        if version.parse(self.api_version) < version.parse("6.4"):
            username_keyword = 'user'
        else:
            username_keyword = 'username'

        obj = self.generate_cgi_data('user.login', {username_keyword: self.username, 'password': self.password}, no_auth=True)
        result = self.api_request(obj)
        self.auth_token = result['result']  # Store the auth token for later use
        # For Zabbix >= 6.4 the server expects HTTP Bearer and no JSON-RPC auth field
        if version.parse(self.api_version) >= version.parse("6.4"):
            self.session.auth = BearerAuth(self.auth_token)
        self.refresh_authentication = False  # Reset the flag after successful login

    def check_authentication(self):
        try:
            # Build request depending on configured auth method
            if conf.servers[self.get_name()].authentication == 'bearer':
                obj = self.generate_cgi_data('user.checkAuthentication', {'token': self.auth_token}, no_auth=True)
            else:
                obj = self.generate_cgi_data('user.checkAuthentication', {'sessionid': self.auth_token}, no_auth=True)

            result = self.api_request(obj, no_auth=True)

            # Zabbix JSON-RPC: success responses contain 'result', errors contain 'error'.
            # Safely handle both without raising KeyError.
            if isinstance(result, dict) and result.get('error'):
                # Server explicitly reported an auth error
                self.refresh_authentication = True
                return

            res = result.get('result') if isinstance(result, dict) else None

            # Interpret common variants:
            # - {'result': True/False}
            # - {'result': {'authenticated': True/False}}
            if isinstance(res, bool):
                self.refresh_authentication = not res
            elif isinstance(res, dict) and 'authenticated' in res:
                self.refresh_authentication = not bool(res.get('authenticated'))
            else:
                # Unable to determine -> force refresh to be safe
                self.refresh_authentication = True
        except ZabbixError:
            # Treat any transport/API error as needing re-authentication
            self.refresh_authentication = True
        except Exception as e:
            raise RuntimeError(f"Authentication check failed: {str(e)}")

    def generate_cgi_data(self, method, params=None, no_auth=False):
        """
            Generate data for Zabbix API requests
        """
        if params is None:
            params = {}
        data = {
            'jsonrpc': '2.0',
            'method': method,
            'params': params,
            'id': 1
        }
        # Only include auth parameter for Zabbix versions before 6.4
        if not no_auth and self.auth_token and version.parse(self.api_version) < version.parse("6.4"):
            data['auth'] = self.auth_token
        return json.dumps(data)

    def set_zabbix_version(self):
        """
            Set the Zabbix API version and other related attributes
        """
        try:
            obj = self.generate_cgi_data('apiinfo.version', no_auth=True)
            result = self.api_request(obj, no_auth=True)
            self.api_version = result['result']
        except Exception as e:
            raise RuntimeError(f"Failed to set Zabbix version: {str(e)}")

    def _get_status(self):
        """
            Get status from Zabbix Server
        """
        ret = Result()
        # create Nagios items dictionary with to lists for services and hosts
        # every list will contain a dictionary for every failed service/host
        # this dictionary is only temporarily

        # =========================================
        # Service
        # =========================================
        try:
            # Get a list of all issues (AKA tripped triggers)
            # add Pagination
            chunk_size = 200
            # services_ids will contain all trigger ids of active services
            results = self.api_request(
                self.generate_cgi_data('trigger.get',
                                       {
                                           'only_true': True,
                                           'skipDependent': True,
                                           'monitored': True,
                                           'active': True,
                                           'output': ['triggerid']
                                       }) )
            services_ids = results['result']
            services = []
            for i in range(0, len(services_ids), chunk_size):
                results = self.api_request(
                    self.generate_cgi_data('trigger.get',
                                           {
                                               'only_true': True,
                                               'skipDependent': True,
                                               'monitored': True,
                                               'active': True,
                                               'output': ['triggerid', 'description', 'lastchange', 'manual_close'],
                                               # 'expandDescription': True,
                                               # 'expandComment': True,
                                               'triggerids': [trigger['triggerid'] for trigger in services_ids[i:i + chunk_size]],
                                               'selectLastEvent': ['eventid', 'name', 'ns', 'clock', 'acknowledged',
                                                                   'value', 'severity'],
                                               'selectHosts': ["hostid", "host", "name", "status", "available",
                                                               "active_available", "maintenance_status", "maintenance_from"],
                                               'selectItems': ['name', 'lastvalue', 'state', 'lastclock']
                                           }))
                services.extend(results['result'])
            for service in services:
                status_information = ", ".join(
                    [f"{item['name']}: {item['lastvalue']}" for item in service['items']])

                # Add opdata to status information if available (from problem API)
                if 'opdata' in service['lastEvent']:
                    if service['lastEvent']['opdata'] != "":
                        status_information = service['lastEvent']['name'] + " (" + service['lastEvent'][
                            'opdata'] + ")"

                service_obj = GenericService()
                service_obj.name = service['lastEvent']['name']
                service_obj.status = self.statemap.get(service['lastEvent']['severity'], service['lastEvent']['severity'])
                service_obj.last_check = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(max(int(item['lastclock']) for item in service['items'])))
                service_obj.duration = human_readable_duration_from_timestamp(service['lastEvent']['clock'])
                service_obj.status_information = status_information
                service_obj.acknowledged = False if service['lastEvent']['acknowledged'] == '0' else True
                #service_obj.address = ''  # Todo: check if address is available
                service_obj.triggerid = service['triggerid']
                service_obj.eventid = service['lastEvent']['eventid']
                service_obj.allow_manual_close = False if str(service['manual_close']) == '0' else True
                scheduled_downtime = False
                if service['hosts']:
                    for host in service['hosts']:
                        if host['maintenance_status'] == '1':
                            scheduled_downtime = True
                        if host['name'] not in self.new_hosts:
                            self.new_hosts[host['name']] = GenericHost()
                            self.new_hosts[host['name']].name = host['name']
                            self.new_hosts[host['name']].server = self.name
                            self.new_hosts[host['name']].status = 'UP'
                            self.new_hosts[host['name']].scheduled_downtime = host['maintenance_status'] == '1'
                        # Map Stuff from Service to Host
                        self.new_hosts[host['name']].services[service['triggerid']] = service_obj
                        self.new_hosts[host['name']].services[service['triggerid']].host = host['name']
                        self.new_hosts[host['name']].services[service['triggerid']].hostid = host['hostid']
                service_obj.scheduled_downtime = scheduled_downtime
        except ZabbixError as e:
            return Result(result=e.result, error=e.result.error)
        except Exception:
            self.isChecking = False
            result, error = self.error(sys.exc_info())
            return Result(result=result, error=error)
        return ret

    def open_monitor(self, host, service_str=""):
        """
            open monitor from treeview context menu
        """
        try:
            host = self.hosts.get(host, None)
            triggerid = None
            for service_obj in host.services.values():
                if service_obj.name == service_str:
                    triggerid = service_obj.triggerid
                    break
            url = f'{self.monitor_url}/zabbix.php?action=problem.view&triggerids%5B%5D={triggerid}&filter_set=1&show_suppressed=1'
            webbrowser_open(url)
        except Exception as e:
            if conf.debug_mode is True:
                self.debug(server=self.get_name(),
                           debug=f'Error while opening monitor page: {e}')
            return

    # Disable set_recheck (nosense in Zabbix)
    def set_recheck(self, info_dict):
        pass

    def _set_acknowledge(self, host, service, author, comment, sticky, notify, persistent, all_services=None):
        if conf.debug_mode is True:
            self.debug(server=self.get_name(), debug=f'Set Acknowledge Host: {host} Service: {service} Sticky: {sticky} persistent: {persistent} All services: {all_services}')
        eventids = set()
        unclosable_events = set()
        if all_services is None:
            all_services = []
        all_services.append(service)
        get_host = self.hosts[host]
        # Through all Services
        for s in all_services:
            # find Trigger ID
            for host_service in get_host.services.values():
                if host_service.name == s:
                    eventid = host_service.eventid
                    # https://github.com/HenriWahl/Nagstamon/issues/826 we may have set eventid = -1 earlier if there was no associated event
                    if eventid == -1:
                        continue
                    eventids.add(eventid)
                    if not host_service.allow_manual_close:
                        unclosable_events.add(eventid)

        # If events pending of acknowledge, execute ack
        if len(eventids) > 0:
            # actions is a bitmask with values:
            # 1 - close problem
            # 2 - acknowledge event
            # 4 - add message
            # 8 - change severity
            # 16 - unacknowledge event
            # 32 - suppress event;
            # 64 - unsuppress event;
            # 128 - change event rank to cause;
            # 256 - change event rank to symptom.
            # sticky = close problem  # TODO: make visible in GUI
            actions = 2
            if comment:
                actions |= 4
            if conf.debug_mode:
                self.debug(server=self.get_name(),
                           debug="Events to acknowledge: " + str(eventids) + " Close: " + str(actions))
            # If some events are not closable, we need to make 2 requests, 1 for the closable and one for the not closable
            if sticky and unclosable_events:
                closable_actions = actions | 1
                closable_events = set(e for e in eventids if e not in unclosable_events)
                self.api_request(
                    self.generate_cgi_data('event.acknowledge',
                                           {
                                               'eventids': list(closable_events),
                                               'message': comment,
                                               'action': closable_actions
                                           }),
                )
                self.api_request(
                    self.generate_cgi_data('event.acknowledge',
                                           {
                                               'eventids': list(unclosable_events),
                                               'message': comment,
                                               'action': actions
                                           }),
                )
            else:
                if sticky:
                    actions |= 1
                try:
                    self.api_request(
                        self.generate_cgi_data('event.acknowledge',
                                               {
                                                   'eventids': list(eventids),
                                                   'message': comment,
                                                   'action': actions
                                               }),
                    )
                except RuntimeError as e:
                    if "Incorrect user name or password or account is temporarily blocked" in str(e):
                        self.error(str(e))
                        return
                    else:
                        raise e

    def _set_downtime(self, hostname, service, author, comment, fixed, start_time, end_time, hours, minutes):
        # Check if there is an associated Application tag with this trigger/item
        triggerid = None
        for host_service in self.hosts[hostname].services:
            if self.hosts[hostname].services[host_service].name == service:
                triggerid = self.hosts[hostname].services[host_service].triggerid
                break
        if self.hosts[hostname].hostid is None:
            self.error("Host ID is None for " + hostname)
            return
        hostids = [self.hosts[hostname].hostid]

        if fixed == 1:
            start_date = datetime.datetime.strptime(start_time, "%Y-%m-%d %H:%M")
            end_date = datetime.datetime.strptime(end_time, "%Y-%m-%d %H:%M")
        else:
            start_date = datetime.datetime.now()
            end_date = start_date + datetime.timedelta(hours=hours, minutes=minutes)

        stime = int(time.mktime(start_date.timetuple()))
        etime = int(time.mktime(end_date.timetuple()))

        if conf.debug_mode is True:
            self.debug(server=self.get_name(), debug=f'Downtime for {hostname}[{hostids}] stime: {stime} etime: {etime}')
        body = {'hostids': hostids, 'name': comment, 'description': author, 'active_since': stime, 'active_till': etime,
                'maintenance_type': 0, "timeperiods": [
                        {"timeperiod_type": 0, "start_date": stime, "period": etime - stime}
                    ]
                }
        if triggerid:
            body['tags'] = [{'tag': 'triggerid', 'operator': 0, 'value': triggerid}]
            body['description'] = f"{body['description']}(Nagstamon): {comment}"
            body['name'] = f'{hostname}: {service}'
        try:
            self.api_request(
                self.generate_cgi_data('maintenance.create', body)
            )
        except ValueError as e:
            if "already exists" in str(e).lower():
                self.debug(server=self.get_name(), debug=f"Maintanence with name {body['name']} already exists")
            else:
                raise e

    def get_start_end(self, host):
        return time.strftime("%Y-%m-%d %H:%M"), time.strftime("%Y-%m-%d %H:%M", time.localtime(time.time() + 7200))

    def get_host(self, host):
        """
            find out ip or hostname of given host to access hosts/devices which do not appear in DNS but
            have their ip saved in Nagios
        """

        # the fasted method is taking hostname as used in monitor
        if conf.connect_by_host is True:
            return Result(result=host)

        ip = ""
        address = host

        try:
            if host in self.hosts:
                ip = self.hosts[host].address
            if conf.debug_mode is True:
                self.debug(server=self.get_name(), host=host, debug=f'IP of {host}: {ip}')

            if conf.connect_by_dns is True:
                try:
                    address = socket.gethostbyaddr(ip)[0]
                except socket.herror:
                    address = ip
            else:
                address = ip
        except ZabbixError:
            result, error = self.error(sys.exc_info())
            return Result(result=result, error=error)

        return Result(result=address)

    def nagiosify_service(self, service):
        """
            next dirty workaround to get Zabbix events to look Nagios-esque
        """
        if ' on ' in service or ' is ' in service:
            for separator in [' on ', ' is ']:
                service = service.split(separator)[0]
        return service



# Nagstamon - Nagios status monitor for your desktop
# Copyright (C) 2008-2026 Henri Wahl <henri@nagstamon.de> et al.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

from Nagstamon.qui.qt import QHBoxLayout


class HBoxLayout(QHBoxLayout):
    """
    Custom QHBoxLayout with zero spacing and margins by default.

    This layout is used to create a horizontal box.

    Args:
        spacing (int, optional): Space between child widgets. Defaults to 0.
        parent (QWidget, optional): Parent widget.
    """

    def __init__(self, spacing=None, parent=None):
        QHBoxLayout.__init__(self, parent)

        if spacing is None:
            self.setSpacing(0)
        else:
            self.setSpacing(spacing)
        # no margin
        self.setContentsMargins(0, 0, 0, 0)



# Nagstamon - Nagios status monitor for your desktop
# Copyright (C) 2008-2026 Henri Wahl <henri@nagstamon.de> et al.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

from copy import deepcopy
from functools import wraps
from urllib.parse import quote

from Nagstamon.servers import SERVER_TYPES
from Nagstamon.config import (Action,
                              conf)
from Nagstamon.qui.dialogs.dialog import Dialog
from Nagstamon.qui.qt import (QMessageBox,
                              Qt,
                              Signal,
                              Slot)


class DialogAction(Dialog):
    """
    Dialog used to set up one single action
    """

    # signal to emit when ok button is pressed - used to update the list of actions
    edited_update_list = Signal(str, str, str)

    # mapping between action types and combobox content
    ACTION_TYPES = {'browser': 'Browser',
                    'command': 'Command',
                    'url': 'URL'}

    def __init__(self):
        Dialog.__init__(self, 'settings_action')
        # initial values
        self.action_conf = None
        self.mode = None
        self.previous_action_conf = None
        # define checkbox-to-widgets dependencies which apply at initialization
        # which widgets have to be hidden because of irrelevance
        # dictionary holds checkbox/radiobutton as key and relevant widgets in a list
        self.TOGGLE_DEPS = {
            self.window.input_checkbox_re_host_enabled: [self.window.input_lineedit_re_host_pattern,
                                                         self.window.input_checkbox_re_host_reverse],
            self.window.input_checkbox_re_service_enabled: [self.window.input_lineedit_re_service_pattern,
                                                            self.window.input_checkbox_re_service_reverse],
            self.window.input_checkbox_re_status_information_enabled: [
                self.window.input_lineedit_re_status_information_pattern,
                self.window.input_checkbox_re_status_information_reverse],
            self.window.input_checkbox_re_duration_enabled: [self.window.input_lineedit_re_duration_pattern,
                                                             self.window.input_checkbox_re_duration_reverse],
            self.window.input_checkbox_re_attempt_enabled: [self.window.input_lineedit_re_attempt_pattern,
                                                            self.window.input_checkbox_re_attempt_reverse],
            self.window.input_checkbox_re_groups_enabled: [self.window.input_lineedit_re_groups_pattern,
                                                           self.window.input_checkbox_re_groups_reverse]}

        # fill action types into combobox
        self.window.input_combobox_type.addItems(sorted(self.ACTION_TYPES.values()))

        # fill default order fields combobox with monitor server types
        self.window.input_combobox_monitor_type.addItem("All monitor servers")
        self.window.input_combobox_monitor_type.addItems(sorted(SERVER_TYPES.keys(), key=str.lower))
        # default to Nagios as it is the mostly used monitor server
        self.window.input_combobox_monitor_type.setCurrentIndex(0)

    def dialog_decoration(method, *args):
        """
        try with a decorator instead of repeated calls
        """

        # the function which decorates method
        # wraps is used to keep the original method's name and docstring
        @wraps(method)
        def decoration_function(self, *decorated_args):
            """
            self.server_conf has to be set by decorated method
            """

            # previous action conf only useful when editing - defaults to None
            self.previous_action_conf = None

            # call decorated method
            method(self, *decorated_args)

            # run through all input widgets and apply defaults from config
            for widget in self.window.__dict__:
                if widget.startswith('input_'):
                    if widget.startswith('input_checkbox_'):
                        setting = widget.split('input_checkbox_')[1]
                        self.window.__dict__[widget].setChecked(self.action_conf.__dict__[setting])
                    elif widget.startswith('input_radiobutton_'):
                        setting = widget.split('input_radiobutton_')[1]
                        self.window.__dict__[widget].setChecked(self.action_conf.__dict__[setting])
                    elif widget.startswith('input_lineedit_'):
                        setting = widget.split('input_lineedit_')[1]
                        self.window.__dict__[widget].setText(self.action_conf.__dict__[setting])
                    elif widget.startswith('input_textedit_'):
                        setting = widget.split('input_textedit_')[1]
                        self.window.__dict__[widget].setText(self.action_conf.__dict__[setting])

            # set comboboxes
            self.window.input_combobox_type.setCurrentText(self.ACTION_TYPES[self.action_conf.type.lower()])
            self.window.input_combobox_monitor_type.setCurrentText(self.action_conf.monitor_type)

            # apply toggle-dependencies between checkboxes and certain widgets
            self.toggle_toggles()

            # important final size adjustment
            self.window.adjustSize()

            # if running on macOS with disabled dock icon, the dock icon might have to be made visible
            # to make Nagstamon accept keyboard input
            self.check_macos_dock_icon_fix_show.emit()

            # force modality - using .open() still allows to switch to main settings dialog
            self.window.exec()

            # en reverse the dock icon might be hidden again after a potential keyboard input
            self.check_macos_dock_icon_fix_hide.emit()

        # give back decorated function
        return decoration_function

    @Slot()
    @dialog_decoration
    def new(self):
        """
        create new server
        """
        self.mode = 'new'

        # create a new server config object
        self.action_conf = Action()
        # window title might be pretty simple
        self.window.setWindowTitle('New action')

    @Slot(str)
    @dialog_decoration
    def edit(self, name):
        """
        edit existing action
        """
        self.mode = 'edit'
        # shorter action conf
        self.action_conf = conf.actions[name]
        # store action name in case it will be changed
        self.previous_action_conf = deepcopy(self.action_conf)
        # set window title
        self.window.setWindowTitle(f'Edit {self.action_conf.name}')

    @Slot(str)
    @dialog_decoration
    def copy(self, name):
        """
        copy existing action
        """
        self.mode = 'copy'
        # shorter action conf
        self.action_conf = deepcopy(conf.actions[name])
        # set window title before name change to reflect copy
        self.window.setWindowTitle(f'Copy {self.action_conf.name}')
        # indicate copy of other action
        self.action_conf.name = f'Copy of {self.action_conf.name}'

    def ok(self):
        """
        evaluate the state of widgets to get new configuration
        """
        # check that no duplicate name exists
        if self.window.input_lineedit_name.text() in conf.actions and \
                (self.mode in ['new', 'copy'] or
                 self.mode == 'edit' and self.action_conf != conf.actions[self.window.input_lineedit_name.text()]):
            # cry if duplicate name exists
            QMessageBox.critical(self.window, 'Nagstamon',
                                 f'The action name <b>{self.window.input_lineedit_name.text()}</b> is already used.',
                                 QMessageBox.StandardButton.Ok)
        else:
            # get configuration from UI
            for widget in self.window.__dict__:
                if widget.startswith('input_'):
                    if widget.startswith('input_checkbox_'):
                        setting = widget.split('input_checkbox_')[1]
                        self.action_conf.__dict__[setting] = self.window.__dict__[widget].isChecked()
                    if widget.startswith('input_radiobutton_'):
                        setting = widget.split('input_radiobutton_')[1]
                        self.action_conf.__dict__[setting] = self.window.__dict__[widget].isChecked()
                    elif widget.startswith('input_combobox_'):
                        setting = widget.split('input_combobox_')[1]
                        self.action_conf.__dict__[setting] = self.window.__dict__[widget].currentText()
                    elif widget.startswith('input_lineedit_'):
                        setting = widget.split('input_lineedit_')[1]
                        self.action_conf.__dict__[setting] = self.window.__dict__[widget].text()
                    elif widget.startswith('input_textedit_'):
                        setting = widget.split('input_textedit_')[1]
                        self.action_conf.__dict__[setting] = self.window.__dict__[widget].toPlainText()

            # edited action will be deleted and recreated with new configuration
            if self.mode == 'edit':
                # delete previous name
                conf.actions.pop(self.previous_action_conf.name)

            # Avoid the wrong monitor type which blocks display of action
            if self.action_conf.monitor_type not in SERVER_TYPES:
                self.action_conf.monitor_type = ''

            # lower type to recognize action type on monitor
            self.action_conf.type = self.action_conf.type.lower()

            # add edited or new/copied action
            conf.actions[self.action_conf.name] = self.action_conf

            # refresh list of actions, give call the current action name to highlight it
            self.edited_update_list.emit('list_actions', 'actions', self.action_conf.name)

            # delete the old action .conf file to reflect name changes
            # new one will be written soon
            if self.previous_action_conf is not None:
                conf.delete_file('actions', 'action_{0}.conf'.format(quote(self.previous_action_conf.name, safe='')))

            # store server settings
            conf.save_multiple_config('actions', 'action')

        # call close and macOS dock icon treatment from ancestor
        super().ok()



# $Id: xtest.py,v 1.1 2000/08/21 10:03:45 petli Exp $
#
# Xlib.ext.composite -- Composite extension module
#
#    Copyright (C) 2007 Peter Liljenberg <peter.liljenberg@gmail.com>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,  USA

"""Composite extension, allowing windows to be rendered to off-screen
storage.

For detailed description, see the protocol specification at
http://freedesktop.org/wiki/Software/CompositeExt

By itself this extension is not very useful, it is intended to be used
together with the DAMAGE and XFIXES extensions.  Typically you would
also need RENDER or glX or some similar method of creating fancy
graphics.
"""

from Xlib import X
from Xlib.protocol import rq
from Xlib.xobject import drawable

extname = 'Composite'

RedirectAutomatic = 0
RedirectManual = 1

class QueryVersion(rq.ReplyRequest):
    _request = rq.Struct(
            rq.Card8('opcode'),
            rq.Opcode(0),
            rq.RequestLength(),
            rq.Card32('major_version'),
            rq.Card32('minor_version')
            )

    _reply = rq.Struct(
            rq.ReplyCode(),
            rq.Pad(1),
            rq.Card16('sequence_number'),
            rq.ReplyLength(),
            rq.Card32('major_version'),
            rq.Card32('minor_version'),
            rq.Pad(16),
            )

def query_version(self):
    return QueryVersion(
        display = self.display,
        opcode = self.display.get_extension_major(extname),
        )


class RedirectWindow(rq.Request):
    _request = rq.Struct(
        rq.Card8('opcode'),
        rq.Opcode(1),
        rq.RequestLength(),
        rq.Window('window'),
        rq.Set('update', 1, (RedirectAutomatic, RedirectManual)),
        rq.Pad(3),
        )

def redirect_window(self, update):
    """Redirect the hierarchy starting at this window to off-screen
    storage.
    """
    RedirectWindow(display = self.display,
                   opcode = self.display.get_extension_major(extname),
                   window = self,
                   update = update,
                   )


class RedirectSubwindows(rq.Request):
    _request = rq.Struct(
        rq.Card8('opcode'),
        rq.Opcode(2),
        rq.RequestLength(),
        rq.Window('window'),
        rq.Set('update', 1, (RedirectAutomatic, RedirectManual)),
        rq.Pad(3),
        )

def redirect_subwindows(self, update):
    """Redirect the hierarchies starting at all current and future
    children to this window to off-screen storage.
    """
    RedirectSubwindows(display = self.display,
                       opcode = self.display.get_extension_major(extname),
                       window = self,
                       update = update,
                       )


class UnredirectWindow(rq.Request):
    _request = rq.Struct(
        rq.Card8('opcode'),
        rq.Opcode(3),
        rq.RequestLength(),
        rq.Window('window'),
        rq.Set('update', 1, (RedirectAutomatic, RedirectManual)),
        rq.Pad(3),
        )

def unredirect_window(self, update):
    """Stop redirecting this window hierarchy.
    """
    UnredirectWindow(display = self.display,
                     opcode = self.display.get_extension_major(extname),
                     window = self,
                     update = update,
                     )


class UnredirectSubindows(rq.Request):
    _request = rq.Struct(
        rq.Card8('opcode'),
        rq.Opcode(4),
        rq.RequestLength(),
        rq.Window('window'),
        rq.Set('update', 1, (RedirectAutomatic, RedirectManual)),
        rq.Pad(3),
        )

def unredirect_subwindows(self, update):
    """Stop redirecting the hierarchies of children to this window.
    """
    RedirectWindow(display = self.display,
                   opcode = self.display.get_extension_major(extname),
                   window = self,
                   update = update,
                   )


class CreateRegionFromBorderClip(rq.Request):
    _request = rq.Struct(
        rq.Card8('opcode'),
        rq.Opcode(5),
        rq.RequestLength(),
        rq.Card32('region'), # FIXME: this should be a Region from XFIXES extension
        rq.Window('window'),
        )

def create_region_from_border_clip(self):
    """Create a region of the border clip of the window, i.e. the area
    that is not clipped by the parent and any sibling windows.
    """
    
    rid = self.display.allocate_resource_id()
    CreateRegionFromBorderClip(
        display = self.display,
        opcode = self.display.get_extension_major(extname),
        region = rid,
        window = self,
        )

    # FIXME: create Region object and return it
    return rid


class NameWindowPixmap(rq.Request):
    _request = rq.Struct(
        rq.Card8('opcode'),
        rq.Opcode(6),
        rq.RequestLength(),
        rq.Window('window'),
        rq.Pixmap('pixmap'),
        )

def name_window_pixmap(self):
    """Create a new pixmap that refers to the off-screen storage of
    the window, including its border.

    This pixmap will remain allocated until freed whatever happens
    with the window.  However, the window will get a new off-screen
    pixmap every time it is mapped or resized, so to keep track of the
    contents you must listen for these events and get a new pixmap
    after them.
    """

    pid = self.display.allocate_resource_id()
    NameWindowPixmap(display = self.display,
                     opcode = self.display.get_extension_major(extname),
                     window = self,
                     pixmap = pid,
                     )

    cls = self.display.get_resource_class('pixmap', drawable.Pixmap)
    return cls(self.display, pid, owner = 1)
    

def init(disp, info):
    disp.extension_add_method('display',
                              'composite_query_version',
                              query_version)

    disp.extension_add_method('window',
                              'composite_redirect_window',
                              redirect_window)

    disp.extension_add_method('window',
                              'composite_redirect_subwindows',
                              redirect_subwindows)

    disp.extension_add_method('window',
                              'composite_unredirect_window',
                              unredirect_window)

    disp.extension_add_method('window',
                              'composite_unredirect_subwindows',
                              unredirect_subwindows)

    disp.extension_add_method('window',
                              'composite_create_region_from_border_clip',
                              create_region_from_border_clip)

    disp.extension_add_method('window',
                              'composite_name_window_pixmap',
                              name_window_pixmap)



# encoding: utf-8
from Nagstamon.helpers import webbrowser_open
# Nagstamon - Nagios status monitor for your desktop
# Copyright (C) 2008-2026 Henri Wahl <henri@nagstamon.de> et al.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

# Nagstamon plugin for Monitos3 from Freicon GmbH & Co. KG
# Based on the Livestatus plugin from MK
# 2017_06_27
# studo

from Nagstamon.objects import Result
from Nagstamon.objects import GenericHost
from Nagstamon.objects import GenericService
from Nagstamon.servers.Generic import GenericServer
from Nagstamon.config import conf

import logging
# logging.basicConfig( level=logging.INFO )
# logging.basicConfig(filename='nagstamon.log',level=logging.INFO)
log = logging.getLogger('Monitos3')
log.setLevel(logging.INFO)

import re
import json
import socket
import time


def format_timestamp(timestamp):
    """format unix timestamp"""
    ts_tuple = time.localtime(timestamp)
    return time.strftime('%Y-%m-%d %H:%M:%S', ts_tuple)


def duration(timestamp):
    """human representation of a duration"""
    factors = (60 * 60 * 24, 60 * 60, 60, 1)
    result = []
    diff = time.time() - timestamp
    for f in factors:
        x = int(diff / f)
        result.append(x)
        diff = diff - x * f
    return '%02dd %02dh %02dm %02ds' % tuple(result)


def service_to_host(data):
    """create the host data blob from the implicit join data of a service"""
    result = {}
    for key in data.keys():
        if key.startswith('host_'):
            result[key[5:]] = data[key]
    return result


class Monitos3Server(GenericServer):
    """A server running Monitos3 with the MK Livestatus NEB. Tested with Monitos3.7.17"""

    TYPE = 'Monitos3'
    MENU_ACTIONS = ['Monitor', 'Recheck', 'Acknowledge', 'Submit check result', 'Downtime']
    # STATES_MAPPING = {'hosts' : {0 : 'UP', 1 : 'DOWN', 2 : 'UNREACHABLE'}, \
    #                 'services' : {0 : 'OK', 1 : 'WARNING', 2 : 'CRITICAL', 3 : 'UNKNOWN'}}
    # STATES_MAPPING_REV = {'hosts' : { 'UP': 0, 'DOWN': 1, 'UNREACHABLE': 2}, \
    #                 'services' : {'OK': 0, 'WARNING': 1, 'CRITICAL': 2, 'UNKNOWN': 3}}
    BROWSER_URLS = { 'monitor': '$MONITOR$/dashboard', \
                    'hosts': '$MONITOR$/monitoring/list/hosts', \
                    'services': '$MONITOR$/monitoring/list/services', \
                    'history': '$MONITOR$/monitoring/list/eventhistory?timestamp>=-7 days'}


    def init_config(self):
        log.info( time.strftime('%a %H:%M:%S') )
        log.info(self.monitor_url)
        # we abuse the monitor_url for the connection information
        self.address = ('localhost', 6558)
        m = re.match(r'.*?://([^:/]+?)(?::(\d+))?(?:/|$)', self.monitor_url)
        if m:
            host, port = m.groups()
            if not port:
                port = 6558
            else:
                port = int(port)
            self.address = (host, port)
        else:
            log.error('unable to parse monitor_url %s', self.monitor_url)
            self.enable = False

    def init_http(self):
        pass

    def communicate(self, data, response=True):
        buffersize = 2**20
        data.append('')
        data.append('')
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        log.debug('connecting')
        s.connect(self.address)
        s.send('\n'.join(data).encode('utf8'))
        if not response:
            log.debug('no response required, disconnect')
            s.close()
            return ''
        result = bytes()
        line = s.recv(buffersize)
        while len(line) > 0:
            result += line
            line = s.recv(buffersize)
        log.debug('disconnect')
        s.close()
        log.debug('received %d bytes', len(result))
        result = result.decode('utf8')
        return result

    def get(self, table, raw=[], headers={}):
        """send data to livestatus socket, receive result, format as json"""
        data = ['GET %s' % table, ]
        headers['OutputFormat'] = 'json'
        headers['ColumnHeaders'] = 'on'
        for k, v in headers.items():
            data.append('%s: %s' % (k, v))
        for line in raw:
            data.append(line)
        result = self.communicate(data)
        if result:
            return json.loads(result)
        return result

    def command(self, *cmd):
        """execute nagios command via livestatus socket.
        For commands see
        https://old.nagios.org/developerinfo/externalcommands/commandlist.php
        """
        data = []
        ts = str(int(time.time()) + 5)  # current epoch timestamp + 5 seconds
        for line in cmd:
            line = 'COMMAND [TIMESTAMP] ' + line
            data.append(line.replace('TIMESTAMP', ts))
        self.communicate(data, response=False)

    def table(self, data):
        """take a livestatus answer and format it as a table,
        list of dictionaries
        [ {host: 'foo1', service: 'bar1'}, {host: 'foo2', service: 'bar2'} ]
        """
        try:
            header = data[0]
        except IndexError:
            raise StopIteration
        for line in data[1:]:
            yield(dict(zip(header, line)))

    def _get_status(self):
        """fetch any host/service not in OK state
        store the information in self.new_hosts
        applies basic filtering. All additional
        filtering and merging new_hosts to hosts
        is left to nagstamon
        """
        log.debug('in def _get_status')
        self.new_hosts = dict()
        filters = []
        filters.append('Filter: state != 0')  # ignore OK state
        if conf.filter_acknowledged_hosts_services:
            filters.append('Filter: acknowledged != 1')
        # hosts
        data = self.get("hosts", raw=filters)
        for h in self.table(data):
            host = self._create_host(h)
            self.new_hosts[host.name] = host
            # from icinga2
            # self.new_hosts[host_name].real_name = h['host_name']
            # 2017_06_02
            log.info("monitos3_host %s with svid %s, alias %s, dn %s, addr %s is %s", host.name, host.svid, host.alias, host.display_name, host.address, host.status)
            # log.debug("host %s is %s", host.name, host.status)
        # services
        data = self.get("services", raw=filters)
        for s in self.table(data):
            # service are attached to host objects
            """
            2017_07_12_22_29_43
            """
            if s['custom_variables']['_HOST_NAME'] in self.new_hosts:
            # if s['host_name'] in self.new_hosts:
                host = self.new_hosts[s['custom_variables']['_HOST_NAME']]
                log.info("In new_hosts services host %s with svid %s, svc %s, svid %s is %s", host.name, host.svid, service.name, service.svid, service.status)
            else:
                # need to create the host
                # icinga2 adds all host information to the server
                # prefixed with HOST_
                xdata = service_to_host(s)  # any field starting with HOST_
                host = self._create_host(xdata)
                self.new_hosts[host.name] = host
                log.info("In else services host %s with svid %s", host.name, host.svid )
            service = self._create_service(s)
            service.host = host.name
            host.services[service.name] = service
            log.debug("All services: host %s with svid %s, svc %s, svid %s is %s", host.name, host.svid, service.name, service.svid, service.status)
            log.debug("monitos3_svcs host.services are: %s", service )
        return Result()

    def _update_object(self, obj, data):
        """populate the generic fields of obj (GenericHost or GenericService)
        from data."""
        result = obj
        result.server = self.name
        result.last_check = format_timestamp(data['last_check'])
        result.duration = duration(data['last_state_change'])
        result.attempt = data['current_attempt']
        result.status_information = data['plugin_output']
        result.passiveonly = False
        result.notifications_disabled = data['notifications_enabled'] != 1
        result.flapping = data['is_flapping'] == 1
        result.acknowledged = data['acknowledged'] == 1
        result.scheduled_downtime = data['scheduled_downtime_depth'] == 1
        if data['state'] == data['last_hard_state']:
            result.status_type = 'hard'
        else:
            result.status_type = 'soft'
        return result

    def _create_host(self, data):
        """create GenericHost from json data"""
        result = self._update_object(GenericHost(), data)
        result.name = data['custom_variables']['_HOST_NAME']
        result.svid = data['name']
        result.display_name = data['custom_variables']['_HOST_NAME']
        result.alias = data['alias']
        result.address = data['address']
        result.custom_variables = data['custom_variables']
        result.show_name = data['custom_variables']['_HOST_NAME']
        # 2017_06_02
        log.debug('data for obj %s is: %s', result.name, data)
        log.debug("monitos3 host name is %s", result.name)
        log.debug("monitos3 host custom_variables are %s", result.custom_variables)
        log.debug("monitos3 host address is %s", result.address)
        # TODO: fix other host states
        host_states = { 0: 'UP', 1: 'DOWN', 2: 'UNREACHABLE' }
        # host_states = { 0: 'UP', 1: 'DOWN', 2: 'UNREACHABLE', 3: 'OTHER' }
        host_state_list = [ 0, 1, 2 ]
        if data['state'] in host_state_list:
            result.status = host_states[data['state']]
        else:
            result.status = 'OTHER'
        return result

    def _create_service(self, data):
        """create GenericService from json data"""
        result = self._update_object(GenericService(), data)
        result.name = data['custom_variables']['_SERVICE_NAME']
        result.svid = data['display_name']
        result.custom_variables = data['custom_variables']
        # result.name = data['display_name']
        service_states = {0: 'OK', 1: 'WARNING', 2: 'CRITICAL', 3: 'UNKNOWN'}
        service_state_list = [ 0, 1, 2, 3 ]
        if data['state'] in service_state_list:
            result.status = service_states[data['state']]
        else:
            result.status = 'OTHER'
        return result

    def set_recheck(self, info_dict):
        host = info_dict['host']
        log.info('host is: %s', host)
        host_svid = self.hosts[host].svid
        log.info('host_svid is: %s', host_svid)
        service = info_dict['service']
        if service:
            svc_svid = self.hosts[host].services[service].svid
            log.info('svc_svid is: %s', svc_svid)
            log.info('service is: %s', service)
            if self.hosts[host].services[service].is_passive_only():
                return
            cmd = ['SCHEDULE_FORCED_SVC_CHECK', host_svid, svc_svid, 'TIMESTAMP']
            log.info('cmd is: %s', cmd )
        else:
            cmd = ['SCHEDULE_FORCED_HOST_CHECK', host_svid, 'TIMESTAMP']
            log.info('cmd is: %s', cmd )
        self.command(';'.join(cmd))
        # log.debug('recheck cmd is: %s', self.command )

    def set_acknowledge(self, info_dict):
        """
        acknowledge a service or host
        {'author': 'username', 'all_services': ['PING', 'DUMMY'], 'host': 'fritz', 'service': '', 'server': <Nagstamon.servers.Monitos3.Monitos3Server object at 0x000001FDF0781CC0>,
        'acknowledge_all_services': True, 'sticky': True, 'notify': True, 'comment': 'acknowledged', 'persistent': True}
        """
        log.info('called def set_acknowledge')
        log.info('info_dict is: %s', info_dict )
        host = info_dict['host']
        log.info('host is: %s', host)
        host_svid = self.hosts[host].svid
        log.info('host_svid is: %s', host_svid)
        service = info_dict['service']
        if service:
            svc_svid = self.hosts[host].services[service].svid
            log.info('svc_svid is: %s', svc_svid)
            cmd = ['ACKNOWLEDGE_SVC_PROBLEM', host_svid, svc_svid]
        else:
            cmd = ['ACKNOWLEDGE_HOST_PROBLEM', host_svid]
        cmd.extend([
            '2' if info_dict['sticky'] else '1',
            '1' if info_dict['notify'] else '0',
            '1' if info_dict['persistent'] else '0',
            info_dict['author'],
            info_dict['comment'],
        ])
        log.info('ack cmd is: %s', ';'.join(cmd) )
        self.command(';'.join(cmd))

    def set_downtime(self, info_dict):
        log.info('set_downtime not implemented')

    def set_submit_check_result(self, info_dict):
        # INFO:Monitos3:info_dict is: {
        # 'host': 'BPMon', 'performance_data': 'ggggg', 'service': 'BP_test cw', 'state': 'critical',
        # 'comment': 'check result submitted', 'server': <Nagstamon.servers.Monitos3.Monitos3Server object at 0x0000021555881CC0>,
        # 'check_output': 'tesstestus'}
        log.debug( json.dumps( str( info_dict ), sort_keys=True, indent=4) )
        host = info_dict['host']
        log.debug('host is: %s', host)
        host_svid = self.hosts[host].svid
        log.debug('host_svid is: %s', host_svid)
        plugin_output = info_dict['check_output']
        service = info_dict['service']
        if service:
            # PROCESS_SERVICE_CHECK_RESULT;<host_name>;<service_description>;<return_code>;<plugin_output>
            log.debug('service is: %s', service)
            svc_svid = self.hosts[host].services[service].svid
            log.debug('svc_svid is: %s', svc_svid)
            rev_service_states = { 'ok': 0, 'warning': 1, 'critical': 2, 'unknown': 3}
            return_code = rev_service_states[ info_dict[ 'state' ] ]
            cmd = ['PROCESS_SERVICE_CHECK_RESULT', host_svid, svc_svid, str( return_code ), plugin_output]
            log.debug('cmd is: %s', cmd )
        else:
            # PROCESS_HOST_CHECK_RESULT;<host_name>;<status_code>;<plugin_output>
            log.debug('host is: %s', host)
            rev_host_states = { 'up': 0, 'down': 1, 'unreachable': 2, 'unknown': 3 }
            return_code = rev_host_states[ info_dict[ 'state' ] ]
            cmd = ['PROCESS_HOST_CHECK_RESULT', host_svid, str( return_code ), plugin_output]
            log.debug('cmd is: %s', cmd )
        self.command(';'.join(cmd))

    def get_start_end(self, host):
        log.info('get_start_end not implemented')
        return 'n/a', 'n/a'

    def open_monitor(self, host, service=''):
        log.info('open_monitor not implemented. host is %s', host)
        # TODO figure out how to add more config options like socket and weburl
        # line 77

    def open_monitor_webpage(self):
        '''
            open monitor from systray/toparea context menu
        '''
        log.info('trying to implement def open_monitor_webpage')
        log.info('address is: %s', self.address)
        # self.address = ('http://localhost')
        monitos3_url = 'http://172.16.10.102'
        log.info('monitos3_url is: %s', monitos3_url )
        webbrowser_open( monitos3_url )

    # TODO
    # config dialog fields
    # config



# Nagstamon - Nagios status monitor for your desktop
# Copyright (C) 2008-2026 Henri Wahl <henri@nagstamon.de> et al.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

from Nagstamon.qui.qt import (QSvgWidget,
                              QSizePolicy,
                              Signal)
from Nagstamon.qui.widgets.draggables import DraggableWidget


class NagstamonLogo(QSvgWidget, DraggableWidget):
    """
    SVG based logo, used for statusbar and top area logos
    """
    # yell if statusbar is moved
    window_moved = Signal()

    # needed for popup after hover
    mouse_entered = Signal()

    # needed for popup after click
    mouse_pressed = Signal()
    # needed to close window in some configurations
    mouse_released = Signal()
    mouse_released_in_window = Signal()

    def __init__(self, file, width=None, height=None, parent=None):
        QSvgWidget.__init__(self, parent=parent)
        # either filepath or QByteArray for top area logo
        self.load(file)
        self.setSizePolicy(QSizePolicy.Policy.Fixed, QSizePolicy.Policy.Fixed)
        # size needed for small Nagstamon logo in statusbar
        if width is not None and height is not None:
            self.setMinimumSize(width, height)
            self.setMaximumSize(width, height)

    def adjust_size(self, height=None, width=None):
        if width is not None and height is not None:
            self.setMinimumSize(width, height)
            self.setMaximumSize(width, height)



# Nagstamon - Nagios status monitor for your desktop
# Copyright (C) 2008-2026 Henri Wahl <henri@nagstamon.de> et al.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA


class GenericObject:
    """
    template for hosts and services
    """

    def __init__(self):
        self.hostid = ''
        self.name = ''
        self.status = ''
        self.status_information = ''
        # default state is soft, to be changed by status_type check
        self.status_type = ''
        self.last_check = ''
        self.duration = ''
        self.attempt = ''
        self.passiveonly = False
        self.acknowledged = False
        self.notifications_disabled = False
        self.flapping = False
        self.scheduled_downtime = False
        # compress all flags like acknowledged and flapping into one string
        self.host_flags = ''
        self.service_flags = ''
        self.visible = True
        # Checkmk also has site info
        self.site = ''
        # server to be added to hash
        self.server = ''
        # might help in Qt
        self.host = ''
        self.service = ''
        self.dummy_column = ''
        # might be used in Op5Monitor and maybe more if groups are used more widely
        self.groups = ''

    def is_passive_only(self):
        return bool(self.passiveonly)

    def is_flapping(self):
        return bool(self.flapping)

    def has_notifications_disabled(self):
        return bool(self.notifications)

    def is_acknowledged(self):
        return bool(self.acknowledged)

    def is_in_scheduled_downtime(self):
        return bool(self.scheduled_downtime)

    def is_visible(self):
        return bool(self.visible)

    def get_name(self):
        """
        return stringified name
        """
        return str(self.name)

    def get_host_name(self):
        """
        Extracts host name from status item.
        Presentation purpose.
        """
        return ''

    def get_service_name(self):
        """
        Extracts service name from status item.
        Presentation purpose.
        """
        return ''

    def get_hash(self):
        """
        returns hash of event status information - different for host and service thus empty here
        """
        return ''

    def get_columns(self, columns_wanted):
        """
        Yield host/service status information for treeview table columns
        """
        for c in columns_wanted:
            yield str(self.__dict__[c])


class GenericHost(GenericObject):

    """
    one host which is monitored by a Nagios server, gets populated with services
    """

    def __init__(self):
        GenericObject.__init__(self)
        # take all the faulty services on host
        self.services = dict()

    def get_host_name(self):
        return str(self.name)

    def is_host(self):
        """
        decides where to put acknowledged/downtime pixbufs in Liststore for Treeview in Popwin
        """
        return True

    def get_hash(self):
        """
        return hash for event history tracking
        """
        return " ".join((self.server, self.site, self.name, self.status))


class GenericService(GenericObject):

    """
    one service which runs on a host
    """

    def __init__(self):
        GenericObject.__init__(self)
        self.unreachable = False

    def get_host_name(self):
        return str(self.host)

    def get_service_name(self):
        return str(self.name)

    def is_host(self):
        """
        decides where to put acknowledged/downtime pixbufs in Liststore for Treeview in Popwin
        """
        return False

    def get_hash(self):
        """
            return hash for event history tracking
        """
        return " ".join((self.server, self.site, self.host, self.name, self.status))


class Result:
    """
    multipurpose result object, used in servers.Generic.fetch_url()
    """
    result = ''
    error = ''
    status_code = 0

    def __init__(self, **kwds):
        # add all keywords to object, every mode searchs inside for its favorite arguments/keywords
        for k in kwds:
            self.__dict__[k] = kwds[k]
